import sys
import os
#import hybrid
from dotenv import load_dotenv
from math import sin, pi
# from hopp.simulation.technologies import REopt
#from hybrid.solar_source import SolarPlant
#from hopp.simulation.technologies.wind_source import WindPlant
import PySAM.Singleowner as so
from matplotlib import use
import pandas as pd
from hopp.simulation.technologies.sites import SiteInfo
from hopp.simulation.technologies.sites import flatirons_site as sample_site
from hopp.simulation.hybrid_simulation import HybridSimulation
from hopp.utilities.log import hybrid_logger as logger
from hopp.utilities.keys import set_developer_nrel_gov_key
from greenheart.to_organize.H2_Analysis.simple_dispatch import SimpleDispatch
from greenheart.simulation.technologies.hydrogen.electrolysis.run_h2_PEM import run_h2_PEM
import numpy as np
from lcoe.lcoe import lcoe as lcoe_calc
import matplotlib.pyplot as plt
from hopp.tools.analysis import create_cost_calculator
import json

import warnings
warnings.filterwarnings("ignore")


def setup_optimize(scenario,wind_size_mw,solar_size_mw,storage_size_mwh,storage_size_mw,solar_cost_multiplier=1.0):
    # Set API key
    load_dotenv()
    NREL_API_KEY = os.getenv("NREL_API_KEY")
    set_developer_nrel_gov_key(NREL_API_KEY)  # Set this key manually here if you are not setting it using the .env

    # Step 1: Establish output structure and special inputs
    # save_all_runs = pd.DataFrame()
    year = 2013
    sample_site['year'] = year
    useful_life = 30
    custom_powercurve = True
    grid_connected_hopp = True
    interconnection_size_mw = 150
    electrolyzer_size = 50000
    kw_continuous = electrolyzer_size*1000

    sample_site['lat'] = scenario['Lat']
    sample_site['lon'] = scenario['Long']
    tower_height = scenario['Tower Height']
    wind_cost_kw = scenario['Wind Cost KW']
    solar_cost_kw = scenario['Solar Cost KW']*solar_cost_multiplier
    storage_cost_kw = scenario['Storage Cost kW']
    storage_cost_kwh = scenario['Storage Cost kWh']

    #Todo: Add useful life to .csv scenario input instead
    scenario['Useful Life'] = useful_life

    site = SiteInfo(sample_site, hub_height=tower_height)

    if storage_size_mw != 0:
        storage_hours = float(storage_size_mwh)/float(storage_size_mw)
    else:
        storage_hours = 0

    technologies = {'solar': solar_size_mw,  # mw system capacity
                    'wind': wind_size_mw,  # mw system capacity
                    'collection_system': True}

    # Create model
    if not grid_connected_hopp:
        interconnection_size_mw = kw_continuous / 1000

    hybrid_plant = HybridSimulation(technologies, site, scenario['Rotor Diameter'], scenario['Tower Height'],
                                    interconnect_kw=electrolyzer_size * 1000,
                                    storage_kw=storage_size_mw * 1000,
                                    storage_kwh=storage_size_mwh * 1000,
                                    storage_hours=storage_hours)

    hybrid_plant.setup_cost_calculator(create_cost_calculator(interconnection_size_mw,
                                                              bos_cost_source='CostPerMW',
                                                              wind_installed_cost_mw=wind_cost_kw * 1000,
                                                              solar_installed_cost_mw=solar_cost_kw * 1000,
                                                              storage_installed_cost_mw=storage_cost_kw * 1000,
                                                              storage_installed_cost_mwh=storage_cost_kwh * 1000,
                                                              storage_hours=storage_hours
                                                              ))
    hybrid_plant.wind.system_model.Turbine.wind_resource_shear = 0.33   
    if solar_size_mw > 0:
        hybrid_plant.solar.financial_model.FinancialParameters.analysis_period = scenario['Useful Life']
        hybrid_plant.solar.financial_model.FinancialParameters.debt_percent = scenario['Debt Equity']
        if scenario['ITC Available']:
            hybrid_plant.solar.financial_model.TaxCreditIncentives.itc_fed_percent = 26
        else:
            hybrid_plant.solar.financial_model.TaxCreditIncentives.itc_fed_percent = 0

    if 'wind' in technologies:
        hybrid_plant.wind.financial_model.FinancialParameters.analysis_period = scenario['Useful Life']
        hybrid_plant.wind.financial_model.FinancialParameters.debt_percent = scenario['Debt Equity']
        if scenario['PTC Available']:
            ptc_val = 0.022
        else:
            ptc_val = 0.0

        interim_list = list(
            hybrid_plant.wind.financial_model.TaxCreditIncentives.ptc_fed_amount)
        interim_list[0] = ptc_val
        hybrid_plant.wind.financial_model.TaxCreditIncentives.ptc_fed_amount = tuple(interim_list)
        hybrid_plant.wind.system_model.Turbine.wind_turbine_hub_ht = scenario['Tower Height']

    if custom_powercurve:
        powercurve_file = open(scenario['Powercurve File'])
        powercurve_data = json.load(powercurve_file)
        powercurve_file.close()
        hybrid_plant.wind.system_model.Turbine.wind_turbine_powercurve_windspeeds = \
            powercurve_data['turbine_powercurve_specification']['wind_speed_ms']
        hybrid_plant.wind.system_model.Turbine.wind_turbine_powercurve_powerout = \
            powercurve_data['turbine_powercurve_specification']['turbine_power_output']

    hybrid_plant.solar.system_capacity_kw = solar_size_mw * 1000
    N = hybrid_plant.wind.num_turbines + 1 # need this to be different than it was b/c else HOPP doesn't change the system capacity
    hybrid_plant.wind.system_capacity_by_num_turbines(wind_size_mw * 1000 * N)

    hybrid_plant.ppa_price = 0.05
    

    return hybrid_plant


def calculate_h_lcoe(bat_model,electrolyzer_size,n_turbines,solar_capacity_mw,battery_storage_mwh,battery_charge_rate,battery_discharge_rate,
                        scenarios_df,buy_from_grid=False,sell_to_grid=False,solar_cost_multiplier=1.0):

    for i, s in scenarios_df.iterrows():
        scenario = s

    wind_size_mw = n_turbines*scenario['Turbine Rating']
    hybrid_plant = setup_optimize(scenario,wind_size_mw,solar_capacity_mw,battery_storage_mwh,battery_discharge_rate,solar_cost_multiplier=solar_cost_multiplier) 

    useful_life = scenario["Useful Life"]

    kw_continuous = electrolyzer_size*1000
    load = [kw_continuous for x in
            range(0, 8760)]  # * (sin(x) + pi) Set desired/required load profile for plant

    sample_site['lat'] = scenario['Lat']
    sample_site['lon'] = scenario['Long']
    force_electrolyzer_cost = scenario['Force Electrolyzer Cost']
    if force_electrolyzer_cost:
        forced_electrolyzer_cost = scenario['Electrolyzer Cost KW']


    # Step 4: HOPP run
    # ------------------------- #

    hybrid_plant.simulate(useful_life)

    # HOPP Specific Energy Metrics
    combined_hybrid_power_production_hopp = hybrid_plant.grid.system_model.Outputs.system_pre_interconnect_kwac[0:8759]
    energy_shortfall_hopp = [x - y for x, y in
                             zip(load,combined_hybrid_power_production_hopp)]
    energy_shortfall_hopp = [x if x > 0 else 0 for x in energy_shortfall_hopp]
    lcoe = hybrid_plant.lcoe_real.hybrid
 

    combined_hybrid_curtailment_hopp = [x - y for x, y in
                             zip(combined_hybrid_power_production_hopp,load)]
    combined_hybrid_curtailment_hopp = [x if x > 0 else 0 for x in combined_hybrid_curtailment_hopp]

    # plt.plot(combined_hybrid_power_production_hopp[0:200], label="production")
    # plt.plot(energy_shortfall_hopp[0:200], label="shortfall")
    # plt.plot(combined_hybrid_curtailment_hopp[0:200], label="curtailment")
    # plt.plot(load[0:200], label="load")
    # plt.legend()
    # plt.show()


    # Step 5: Run Simple Dispatch Model
    # # ------------------------- #
    bat_model.Nt = len(combined_hybrid_curtailment_hopp)
    bat_model.curtailment = combined_hybrid_curtailment_hopp
    bat_model.shortfall = energy_shortfall_hopp

    bat_model.battery_storage = battery_storage_mwh * 1000
    bat_model.charge_rate = battery_charge_rate * 1000
    bat_model.discharge_rate = battery_discharge_rate * 1000

    battery_used, excess_energy, battery_SOC = bat_model.run()
    combined_pv_wind_storage_power_production_hopp = combined_hybrid_power_production_hopp + battery_used
    energy_to_electrolyzer = [x if x < kw_continuous else kw_continuous for x in combined_pv_wind_storage_power_production_hopp]

    if sell_to_grid:
        profit_from_selling_to_grid = np.sum(excess_energy)*sell_to_grid
    else:
        profit_from_selling_to_grid = 0.0

    if buy_from_grid:
        cost_to_buy_from_grid = 0.0
        
        for i in range(len(combined_pv_wind_storage_power_production_hopp)):
            if combined_pv_wind_storage_power_production_hopp[i] < kw_continuous:
                cost_to_buy_from_grid += (kw_continuous-combined_pv_wind_storage_power_production_hopp[i])*buy_from_grid
                combined_pv_wind_storage_power_production_hopp[i] = kw_continuous
    else:
        cost_to_buy_from_grid = 0.0

    # Step 6: Run the Python H2A model
    # ------------------------- #
    electrical_generation_timeseries = np.zeros_like(energy_to_electrolyzer)
    electrical_generation_timeseries[:] = energy_to_electrolyzer[:]
    
    adjusted_installed_cost = hybrid_plant.grid.financial_model.Outputs.adjusted_installed_cost
    net_capital_costs = hybrid_plant.grid.financial_model.SystemCosts.total_installed_cost

    H2_Results, H2A_Results = run_h2_PEM(electrical_generation_timeseries,electrolyzer_size,
                    kw_continuous,forced_electrolyzer_cost,lcoe,adjusted_installed_cost,useful_life,
                    net_capital_costs)

    # print(H2_Results['hydrogen_annual_output'])
    total_hopp_installed_cost = hybrid_plant.grid.financial_model.SystemCosts.total_installed_cost
    total_electrolyzer_cost = H2A_Results['scaled_total_installed_cost']
    total_system_installed_cost = total_hopp_installed_cost + total_electrolyzer_cost
    annual_operating_cost_hopp = (wind_size_mw * 1000 * 42) + (solar_capacity_mw * 1000 * 13)
    annual_operating_cost_h2 = H2A_Results['Fixed O&M'] * H2_Results['hydrogen_annual_output']
    total_annual_operating_costs = annual_operating_cost_hopp + annual_operating_cost_h2 + cost_to_buy_from_grid - profit_from_selling_to_grid

    h_lcoe = lcoe_calc((H2_Results['hydrogen_annual_output']), total_system_installed_cost,
                    total_annual_operating_costs, 0.07, useful_life)

    return h_lcoe, np.sum(combined_hybrid_power_production_hopp), H2_Results['hydrogen_annual_output'], total_system_installed_cost, total_annual_operating_costs


def setup_power_calcs(scenario,wind_size_mw,solar_size_mw,storage_size_mwh,storage_size_mw):
    # Set API key
    load_dotenv()
    NREL_API_KEY = os.getenv("NREL_API_KEY")
    set_developer_nrel_gov_key(NREL_API_KEY)  # Set this key manually here if you are not setting it using the .env

    # Step 1: Establish output structure and special inputs
    # save_all_runs = pd.DataFrame()
    year = 2013
    sample_site['year'] = year
    useful_life = 30
    custom_powercurve = True
    electrolyzer_size = 50000

    sample_site['lat'] = scenario['Latitude']
    sample_site['lon'] = scenario['Longitude']
    tower_height = scenario['Tower Height']

    #Todo: Add useful life to .csv scenario input instead
    scenario['Useful Life'] = useful_life

    site = SiteInfo(sample_site, hub_height=tower_height)

    if storage_size_mw != 0:
        storage_hours = float(storage_size_mwh)/float(storage_size_mw)
    else:
        storage_hours = 0

    technologies = {'solar': solar_size_mw,  # mw system capacity
                    'wind': wind_size_mw,  # mw system capacity
                    'collection_system': True}

    hybrid_plant = HybridSimulation(technologies, site, scenario['Rotor Diameter'], scenario['Tower Height'],
                                    interconnect_kw=electrolyzer_size * 1000,
                                    storage_kw=storage_size_mw * 1000,
                                    storage_kwh=storage_size_mwh * 1000,
                                    storage_hours=storage_hours)

    hybrid_plant.wind.system_model.Turbine.wind_resource_shear = 0.33   

    if custom_powercurve:
        powercurve_file = open(scenario['Power Curve File'])
        powercurve_data = json.load(powercurve_file)
        powercurve_file.close()
        hybrid_plant.wind.system_model.Turbine.wind_turbine_powercurve_windspeeds = \
            powercurve_data['turbine_powercurve_specification']['wind_speed_ms']
        hybrid_plant.wind.system_model.Turbine.wind_turbine_powercurve_powerout = \
            powercurve_data['turbine_powercurve_specification']['turbine_power_output']

    hybrid_plant.solar.system_capacity_kw = solar_size_mw * 1000
    N = hybrid_plant.wind.num_turbines + 1 # need this to be different than it was b/c else HOPP doesn't change the system capacity
    hybrid_plant.wind.system_capacity_by_num_turbines(wind_size_mw * 1000 * N)    

    return hybrid_plant


def setup_cost_calcs(scenario,hybrid_plant,electrolyzer_size_mw,wind_size_mw,solar_size_mw,
                    storage_size_mwh,storage_size_mw,solar_cost_multiplier=1.0):

    # Step 1: Establish output structure and special inputs
    # save_all_runs = pd.DataFrame()
    year = 2013
    sample_site['year'] = year
    useful_life = 30
    # grid_connected_hopp = True
    # interconnection_size_mw = 150
    # electrolyzer_size = 50000
    # kw_continuous = electrolyzer_size*1000

    sample_site['lat'] = scenario['Latitude']
    sample_site['lon'] = scenario['Longitude']
    tower_height = scenario['Tower Height']
    wind_cost_kw = scenario['Wind Cost (kW)']
    solar_cost_kw = scenario['Solar Cost (kW)']*solar_cost_multiplier
    storage_cost_kw = scenario['Storage Cost (kW)']
    storage_cost_kwh = scenario['Storage Cost (kWh)']

    #Todo: Add useful life to .csv scenario input instead
    scenario['Useful Life'] = useful_life

    if storage_size_mw != 0:
        storage_hours = float(storage_size_mwh)/float(storage_size_mw)
    else:
        storage_hours = 0

    # Create model
    # if not grid_connected_hopp:
    #     interconnection_size_mw = kw_continuous / 1000

    interconnection_size_mw = electrolyzer_size_mw
    # print(interconnection_size_mw)

    hybrid_plant.setup_cost_calculator(create_cost_calculator(interconnection_size_mw,
                                                              bos_cost_source='CostPerMW',
                                                              wind_installed_cost_mw=wind_cost_kw * 1000,
                                                              solar_installed_cost_mw=solar_cost_kw * 1000,
                                                              storage_installed_cost_mw=storage_cost_kw * 1000,
                                                              storage_installed_cost_mwh=storage_cost_kwh * 1000,
                                                              storage_hours=storage_hours
                                                              ))
    hybrid_plant.wind.system_model.Turbine.wind_resource_shear = 0.33   
    if solar_size_mw > 0:
        hybrid_plant.solar.financial_model.FinancialParameters.analysis_period = scenario['Useful Life']
        hybrid_plant.solar.financial_model.FinancialParameters.debt_percent = scenario['Debt Equity']
        if scenario['ITC']:
            hybrid_plant.solar.financial_model.TaxCreditIncentives.itc_fed_percent = 26
        else:
            hybrid_plant.solar.financial_model.TaxCreditIncentives.itc_fed_percent = 0

    if wind_size_mw > 0:
        hybrid_plant.wind.financial_model.FinancialParameters.analysis_period = scenario['Useful Life']
        hybrid_plant.wind.financial_model.FinancialParameters.debt_percent = scenario['Debt Equity']
        if scenario['PTC']:
            ptc_val = 0.022
        else:
            ptc_val = 0.0

        interim_list = list(
            hybrid_plant.wind.financial_model.TaxCreditIncentives.ptc_fed_amount)
        interim_list[0] = ptc_val
        hybrid_plant.wind.financial_model.TaxCreditIncentives.ptc_fed_amount = tuple(interim_list)
        hybrid_plant.wind.system_model.Turbine.wind_turbine_hub_ht = scenario['Tower Height']

    hybrid_plant.ppa_price = 0.05
    hybrid_plant.wind.system_capacity_kw = wind_size_mw*1000
    return hybrid_plant


def calculate_h_lcoe_continuous(bat_model,electrolyzer_size,wind_capacity_mw,solar_capacity_mw,battery_storage_mwh,battery_charge_rate,battery_discharge_rate,
                        scenario,buy_from_grid=False,sell_to_grid=False,solar_cost_multiplier=1.0):

    wind_size_mw = scenario['Turbine Rating']
    hybrid_plant = setup_power_calcs(scenario,wind_size_mw,solar_capacity_mw,battery_storage_mwh,battery_discharge_rate) 

    useful_life = scenario["Useful Life"]

    kw_continuous = electrolyzer_size*1000
    load = [kw_continuous for x in
            range(0, 8760)]  # * (sin(x) + pi) Set desired/required load profile for plant

    sample_site['lat'] = scenario['Latitude']
    sample_site['lon'] = scenario['Longitude']
    force_electrolyzer_cost = scenario['Force Electrolyzer Cost']
    if force_electrolyzer_cost:
        forced_electrolyzer_cost = scenario['Electrolyzer Cost (kW)']


    # Step 4: HOPP run
    # ------------------------- #

    hybrid_plant.simulate_power(useful_life)

    # HOPP Specific Energy Metrics
    pv_power_production = hybrid_plant.solar.generation_profile()[0:8760]
    wind_power_production_1 = hybrid_plant.wind.generation_profile()[0:8760]
    effective_n_turbs = wind_capacity_mw/scenario['Turbine Rating']
    wind_power_production = np.array(wind_power_production_1)*effective_n_turbs/hybrid_plant.wind.num_turbines

    # plt.plot(pv_power_production)
    # plt.plot(wind_power_production)
    # plt.show()

    combined_hybrid_power_production_hopp = pv_power_production + wind_power_production
    energy_shortfall_hopp = [x - y for x, y in
                             zip(load,combined_hybrid_power_production_hopp)]
    energy_shortfall_hopp = [x if x > 0 else 0 for x in energy_shortfall_hopp]
 
    combined_hybrid_curtailment_hopp = [x - y for x, y in
                             zip(combined_hybrid_power_production_hopp,load)]
    combined_hybrid_curtailment_hopp = [x if x > 0 else 0 for x in combined_hybrid_curtailment_hopp]

    # # plt.plot(combined_hybrid_power_production_hopp[0:200], label="production")
    # # plt.plot(energy_shortfall_hopp[0:200], label="shortfall")
    # # plt.plot(combined_hybrid_curtailment_hopp[0:200], label="curtailment")
    # # plt.plot(load[0:200], label="load")
    # # plt.legend()
    # # plt.show()


    # Step 5: Run Simple Dispatch Model
    # # ------------------------- #
    bat_model.Nt = len(combined_hybrid_curtailment_hopp)
    bat_model.curtailment = combined_hybrid_curtailment_hopp
    bat_model.shortfall = energy_shortfall_hopp

    bat_model.battery_storage = battery_storage_mwh * 1000
    bat_model.charge_rate = battery_charge_rate * 1000
    bat_model.discharge_rate = battery_discharge_rate * 1000

    battery_used, excess_energy, battery_SOC = bat_model.run()
    combined_pv_wind_storage_power_production_hopp = combined_hybrid_power_production_hopp + battery_used

    if sell_to_grid:
        profit_from_selling_to_grid = np.sum(excess_energy)*sell_to_grid
    else:
        profit_from_selling_to_grid = 0.0

    if buy_from_grid:
        cost_to_buy_from_grid = 0.0
        
        for i in range(len(combined_pv_wind_storage_power_production_hopp)):
            if combined_pv_wind_storage_power_production_hopp[i] < kw_continuous:
                cost_to_buy_from_grid += (kw_continuous-combined_pv_wind_storage_power_production_hopp[i])*buy_from_grid
                combined_pv_wind_storage_power_production_hopp[i] = kw_continuous
    else:
        cost_to_buy_from_grid = 0.0

    energy_to_electrolyzer = [x if x < kw_continuous else kw_continuous for x in combined_pv_wind_storage_power_production_hopp]
    electrolyzer_CF = np.sum(energy_to_electrolyzer)/(kw_continuous*len(energy_to_electrolyzer))

    # Step 6: Run the Python H2A model
    # ------------------------- #
    hybrid_plant = setup_cost_calcs(scenario,hybrid_plant,electrolyzer_size,wind_capacity_mw,solar_capacity_mw,
                        battery_storage_mwh,battery_charge_rate,solar_cost_multiplier=solar_cost_multiplier)
    hybrid_plant.simulate_costs(combined_hybrid_power_production_hopp, useful_life)
    lcoe = hybrid_plant.lcoe_real.hybrid

    # print("lcoe: ", lcoe)

    electrical_generation_timeseries = np.zeros_like(energy_to_electrolyzer)
    electrical_generation_timeseries[:] = energy_to_electrolyzer[:]
    
    adjusted_installed_cost = hybrid_plant.grid.financial_model.Outputs.adjusted_installed_cost
    net_capital_costs = hybrid_plant.grid.financial_model.SystemCosts.total_installed_cost

    H2_Results, H2A_Results = run_h2_PEM(electrical_generation_timeseries,electrolyzer_size,
                    kw_continuous,forced_electrolyzer_cost,lcoe,adjusted_installed_cost,useful_life,
                    net_capital_costs)

    # print(H2_Results['hydrogen_annual_output'])
    total_hopp_installed_cost = hybrid_plant.grid.financial_model.SystemCosts.total_installed_cost
    total_electrolyzer_cost = H2A_Results['scaled_total_installed_cost']
    total_system_installed_cost = total_hopp_installed_cost + total_electrolyzer_cost
    annual_operating_cost_hopp = (wind_size_mw * 1000 * 42) + (solar_capacity_mw * 1000 * 13)
    annual_operating_cost_h2 = H2A_Results['Fixed O&M'] * H2_Results['hydrogen_annual_output']
    total_annual_operating_costs = annual_operating_cost_hopp + annual_operating_cost_h2 + cost_to_buy_from_grid - profit_from_selling_to_grid

    h_lcoe = lcoe_calc((H2_Results['hydrogen_annual_output']), total_system_installed_cost,
                    total_annual_operating_costs, 0.07, useful_life)

    return h_lcoe, np.sum(combined_hybrid_power_production_hopp), H2_Results['hydrogen_annual_output'], total_system_installed_cost, total_annual_operating_costs,electrolyzer_CF


if __name__=="__main__":
    bat_model = SimpleDispatch()

    electrolyzer_size_mw = 500
    solar_capacity_mw = 500
    wind_capacity_mw = 500
    battery_storage_mwh = 500
    battery_storage_mw = 0

    N = 20
    h_lcoe = np.zeros(N)
    pv_wind = np.zeros(N)
    hydrogen_annual_output = np.zeros(N)
    total_installed_cost = np.zeros(N)
    total_operating_cost = np.zeros(N)
    CF = np.zeros(N)

    # n_turbines = np.linspace(10,100,N)
    battery_storage_mwh = np.linspace(0.0,1000.0,N)

    scenarios_df = pd.read_csv('single_scenario2.csv') 
    for i, s in scenarios_df.iterrows():
        scenario = s

    buy_from_grid = False
    for i in range(N):
        h_lcoe[i], pv_wind[i], hydrogen_annual_output[i], total_installed_cost[i], total_operating_cost[i],CF[i] = \
                        calculate_h_lcoe_continuous(bat_model,electrolyzer_size_mw,wind_capacity_mw,
                        solar_capacity_mw,battery_storage_mwh[i],battery_storage_mw,battery_storage_mw,
                        scenario,buy_from_grid=buy_from_grid,sell_to_grid=False)

    print("h_lcoe: ", h_lcoe)
    print("pv_wind: ", pv_wind)
    print("h2 output: ", hydrogen_annual_output)
    print("total_installed_cost: ", total_installed_cost)
    print("total_operating_cost: ", total_operating_cost)


    metric = battery_storage_mwh
    axis_label=("battery_storage_mwh")

    plt.figure(figsize=(8,6))
    plt.subplot(321)
    plt.plot(metric,h_lcoe,"o")
    plt.xlabel(axis_label)
    plt.title("h_lcoe")

    plt.subplot(322)
    plt.plot(metric,CF,"o")
    plt.xlabel(axis_label)
    plt.title("electrolyzer capacity factor")

    plt.subplot(323)
    plt.plot(metric,hydrogen_annual_output,"o")
    plt.xlabel(axis_label)
    plt.title("hydrogen_annual_output")

    plt.subplot(324)
    plt.plot(metric,total_installed_cost,"o")
    plt.xlabel(axis_label)
    plt.title("total_installed_cost")

    plt.subplot(325)
    plt.plot(metric,total_operating_cost,"o")
    plt.xlabel(axis_label)
    plt.title("total_operating_cost")

    plt.subplot(326)
    plt.plot(metric,pv_wind,"o")
    plt.xlabel(axis_label)
    plt.title("pv_wind")

    plt.tight_layout()
    plt.show()


"""
Author:Charles Kiefer
Date: 4/11/2023
Institution: National Renewable Energy Lab 
Description: This file shall handle costing and sizing of offshore floating platforms deicated to hydrogen production.  It uses the 
             same foundation as fixed_platform.py.  Both have been modeled off of existing BOS cost/sizing calculations fond in ORBIT.
             It can be run as standalone functions or as appended ORBIT project phases.


             
Sources:
    - [1] ORBIT: https://github.com/WISDEM/ORBIT v1.1
Args:
    - tech_required_area: (float): area needed for combination of all tech (m^2), not including buffer or working space
    - tech_combined_mass: (float): mass of all tech being placed on the platform (kg or tonnes)year

   
    - depth: (float): bathometry at the platform location (m) ##Site depths for floating projects need to be at depths 500 m to 1500 m because of Orbit Semitaut limitations (7/31)
    - distance_to_port: (float): distance ships must travel from port to site location (km)
    
    Future arguments: (Not used at this time)
    - construction year  (int): 
    - lifetime (int): lifetime of the plant in years (may not be needed)

Returns:
    - platform_mass (float): Adjusted mass of platform + substructure
    - design_capex (float): capital expenditures (platform design + substructure fabrication)
    - installation_capex (float): capital expenditures (installation cost)
    - platform_opex (float): the OPEX (annual, fixed) in USD for the platform

"""
''' 
Notes:
    Thank you Jake Nunemaker's oswh2 repository and Rebecca Fuchs SemiTaut_mooring repository!!!
    pile_cost=0 $US/tonne for monopile construction. Not a bug, this # is consistent with the rest of ORBIT
'''

import os
import math
# 
from ORBIT import ProjectManager, load_config
from ORBIT.core import Vessel
from ORBIT.core.library import initialize_library
from ORBIT.phases.design import DesignPhase
from ORBIT.phases.install import InstallPhase
from ORBIT.phases.design import MooringSystemDesign

from scipy.interpolate import interp1d
import numpy as np

from greenheart.simulation.technologies.offshore.all_platforms import calc_platform_opex, install_platform

class FloatingPlatformDesign(DesignPhase):
    '''
    This is a modified class based on ORBIT's design phase
    '''

    #phase = "H2 Floating Platform Design"

    # Expected inputs from config yaml file
    expected_config = {
        "site": {
            "distance" : "int | float",
            "depth" : "int | float",
        }, 

        "equipment": {
            "tech_required_area" : "float", 
            "tech_combined_mass" : "float",
            "topside_design_cost": "USD (optional, default:4.5e6)",
            "fabrication_cost_rate": "USD/t (optional, default: 14500.)",
            "substructure_steel_rate": "USD/t (optional, default: 3000.)",
        }

    }

    # Takes in arguments and initialize library files
    def __init__(self, config, **kwargs):
        
        self.phase = "H2 Floating Platform Design"

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)

        self._outputs = {}
            # Runs the design cost models 

    def run(self):
        
        #print("Floating Platform Design run() is working!!!")

        self.distance = self.config['site']['distance']     # km
        self.depth = self.config['site']['depth']           # m

        _platform = self.config.get('equipment',{})

        self.mass = _platform.get('tech_combined_mass',999)     # t
        self.area = _platform.get('tech_required_area', 1000)   # m**2

        design_cost = _platform.get('topside_design_cost', 4.5e6)   # USD
        fab_cost_rate = _platform.get('fabrication_cost_rate', 14500.)   # USD/t
        steel_cost = _platform.get('substructure_steel_rate', 3000) # USD/t
        ##NEED updated version
        # Add individual calcs/functions in the run() method
        '''Calls in SemiTaut Costs and Variables for Substructure mass and cost'''
        self.anchor_type = "Drag Embedment"
        self.mooring_type = "Semitaut"
        self.num_lines = 4
        MooringSystemDesign.MooringSystemDesign.calculate_line_length_mass(self)
        MooringSystemDesign.MooringSystemDesign.calculate_anchor_mass_cost(self)
        MooringSystemDesign.MooringSystemDesign.determine_mooring_line_cost(self)
        total_cost, total_mass = calc_substructure_mass_and_cost(self.mass, self.area, 
                        self.depth, fab_cost_rate, design_cost, steel_cost,
                        self.line_cost, self.anchor_cost, self.anchor_mass, self.line_mass, self.num_lines)

        # Create an ouput dict 
        self._outputs['floating_platform'] = {
            "mass" : total_mass, 
            "area" : self.area,
            "total_cost" : total_cost
        }

    # A design object needs to have attribute design_result and detailed_output
    @property
    def design_result(self):

        return {
            "platform_design":{
                "mass" : self._outputs['floating_platform']['mass'],
                "area" : self._outputs['floating_platform']['area'],
                "total_cost": self._outputs['floating_platform']['total_cost'],
            }
        }

    @property
    def detailed_output(self):

        return {}

class FloatingPlatformInstallation(InstallPhase):
    '''
    This is a modified class based on ORBIT's install phase 
    '''

    #phase = "H2 Floating Platform Installation"
    
    # Expected inputs from config yaml file
    expected_config = {
        "site": {
            "distance" : "int | float",
            "depth" : "int | float",
        }, 

        "equipment": {
            "tech_required_area" : "float", 
            "tech_combined_mass" : "float",
            "install_duration": "days (optional, default: 14)",
        },

        "oss_install_vessel" : "str | dict",
    }

    # Need to initialize arguments and weather files 
    def __init__(self, config, weather=None, **kwargs):
        
        super().__init__(weather, **kwargs)

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)

        self.initialize_port()
        self.setup_simulation(**kwargs)

    # Setup simulation seems to be the install phase's equivalent run() module
    def setup_simulation(self, **kwargs):

        #print("Floating Platform Install setup_sim() is working!!!")

        self.distance = self.config['site']['distance']
        self.depth = self.config['site']['depth']
        self.mass = self.config['equipment']['tech_combined_mass']
        self.area = self.config['equipment']['tech_required_area']

        _platform = self.config.get('equipment', {})
        design_cost = _platform.get('topside_design_cost', 4.5e6)   # USD
        fab_cost_rate = _platform.get('fabrication_cost_rate', 14500.)   # USD/t
        steel_cost = _platform.get('substructure_steel_rate', 3000) # USD/t
        
        install_duration = _platform.get("install_duration", 14)    # days
        
        # Initialize vessel 
        vessel_specs = self.config.get("oss_install_vessel", None)
        name = vessel_specs.get("name","Offshore Substation Install Vessel")

        vessel = Vessel(name, vessel_specs)
        self.env.register(vessel)

        vessel.initialize()
        self.install_vessel = vessel
        
        # Add in the mass of the substructure to total mass (may or may not impact the final install cost)

        '''Calls in SemiTaut Costs and Variables'''
        self.anchor_type = "Drag Embedment"
        self.mooring_type = "Semitaut"
        self.num_lines = 4
        MooringSystemDesign.MooringSystemDesign.calculate_line_length_mass(self)
        MooringSystemDesign.MooringSystemDesign.calculate_anchor_mass_cost(self)
        MooringSystemDesign.MooringSystemDesign.determine_mooring_line_cost(self)
        
        _, substructure_mass = calc_substructure_mass_and_cost(self.mass, self.area, 
                        self.depth, fab_cost_rate, design_cost, steel_cost,
                        self.line_cost, self.anchor_cost, self.anchor_mass, self.line_mass, self.num_lines)

        total_mass = substructure_mass  # t

         # Call the install_platform function
        self.install_capex = install_platform(total_mass, self.area, self.distance, \
                                                   install_duration, self.install_vessel, foundation="floating")

    # An install object needs to have attribute system_capex, installation_capex, and detailed output
    @property
    def system_capex(self):

        return {}

    @property 
    def installation_capex(self):
        
        return self.install_capex

    @property
    def detailed_output(self):

        return {}

# Define individual calculations and functions to use outside or with ORBIT
def calc_substructure_mass_and_cost(mass, area, depth, fab_cost_rate=14500., design_cost=4.5e6, sub_cost_rate=3000, line_cost=0, anchor_cost=0, anchor_mass=0, line_mass=0,num_lines=4):
    '''
    calc_substructure_mass_and_cost returns the total mass including substructure, topside and equipment.  Also returns the cost of the substructure and topside
    Inputs: mass            | Mass of equipment on platform (tonnes)
            area            | Area needed for equipment (meter^2) (not necessary)
            depth           | Ocean depth at platform location (meters)
            fab_cost_rate   | Cost rate to fabricate topside (USD/tonne)
            design_cost     | Design cost to design structural components (USD) from ORBIT
            sub_cost_rate   | Steel cost rate (USD/tonne) from ORBIT'''

    '''
    Platform is substructure and topside combined
    All functions are based off NREL's ORBIT (oss_design)
    default values are specified in ORBIT
    '''
    topside_mass = mass
    topside_fab_cost_rate   =   fab_cost_rate   
    topside_design_cost     =   design_cost

    '''Topside Cost & Mass
    Topside Mass is the required Mass the platform will hold
    Topside Cost is a function of topside mass, fab cost and design cost'''
    topside_cost   =   topside_mass*topside_fab_cost_rate + topside_design_cost #USD

    '''Substructure
    Substructure Mass is a function of the topside mass
    Substructure Cost is a function of of substructure mass pile mass and cost rates for each'''

    substructure_cost_rate  =   sub_cost_rate        # USD/t

    substructure_mass       =   0.4*topside_mass        # t
    substructure_cost       =   (substructure_mass*substructure_cost_rate)     # USD  
    substructure_total_mass =   substructure_mass       # t

    '''Total Mooring cost and mass for the substructure
    Line_cost, anchor_cost, line_mass, anchor_mass are grabbed from MooringSystemDesign in ORBIT
    Mooring_mass is returned in kilograms and will need to '''
    mooring_cost = (line_cost + anchor_cost)*num_lines #USD
    mooring_mass = (line_mass + anchor_mass)*num_lines #kg
    
    '''Total Platform capex = capex Topside + capex substructure'''
    total_capex = 2*(topside_cost + substructure_cost + mooring_cost)
    platform_capex  = total_capex # USD
    platform_mass   = substructure_total_mass + topside_mass + mooring_mass/1000   # t 
    #mass of equipment and floating substructure for substation
    
    return platform_capex, platform_mass


# Standalone test sections 
if __name__ == '__main__':
    print("\n*** New FloatingPlatform Standalone test section ***\n")

    orbit_libpath = os.path.abspath(os.path.join(os.getcwd(), os.pardir, os.pardir, os.pardir, 'ORBIT', 'library'))
    print(orbit_libpath)
    initialize_library(orbit_libpath)

    config_path = os.path.abspath(__file__)
    config_fname = load_config(os.path.join(config_path, os.pardir, "example_floating_project.yaml"))

    
    #ProjectManager._design_phases.append(FloatingPlatformDesign)
    ProjectManager.register_design_phase(FloatingPlatformDesign)
    #ProjectManager._install_phases.append(FloatingPlatformInstallation)
    ProjectManager.register_install_phase(FloatingPlatformInstallation)

    platform = ProjectManager(config_fname)
    platform.run()

    design_capex = platform.design_results['platform_design']['total_cost']
    install_capex = platform.installation_capex

    #print("Project Params", h2platform.project_params.items())
    platform_opex = calc_platform_opex((design_capex + install_capex))

    print("ORBIT Phases: ", platform.phases.keys())
    print(f"\tH2 Platform Design Capex:    {design_capex:.0f} USD")
    print(f"\tH2 Platform Install Capex:  {install_capex:.0f} USD")
    print('')
    print(f"\tTotal H2 Platform Capex:   {(design_capex+install_capex)/1e6:.0f} mUSD")
    print(f"\tH2 Platform Opex: {platform_opex:.0f} USD/year")



# green steel code

from csv import writer
import os
import sys
sys.path.append("c:/Users/CKIEFER/GreenSteel/HDRI-EAF-Technoeconomic-model/NREL/HOPP")
sys.path.append("C:/Users/CKIEFER/GreenSteel/HDRI-EAF-Technoeconomic-model")
from dotenv import load_dotenv
import pandas as pd
import json
from hopp.simulation.technologies.sites import SiteInfo
from hopp.simulation.technologies.sites import flatirons_site as sample_site
from hopp.utilities.keys import set_developer_nrel_gov_key
from greenheart.to_organize.plot_reopt_results import plot_reopt_results
from greenheart.to_organize.run_reopt import run_reopt
from greenheart.to_organize.run_reopt import run_reopt
from greenheart.to_organize.H2_Analysis.hopp_for_h2 import hopp_for_h2
from greenheart.to_organize.H2_Analysis.hopp_for_h2 import run_h2a as run_h2a #no h2a function
from greenheart.to_organize.H2_Analysis.simple_dispatch import SimpleDispatch
import hopp.simulation.technologies.hydrogen.electrolysis.run_h2_PEM as run_h2_PEM
import numpy as np
import numpy_financial as npf
from lcoe.lcoe import lcoe as lcoe_calc
import matplotlib.pyplot as plt
import warnings
from pathlib import Path
warnings.filterwarnings("ignore")


# Set API key
load_dotenv()
NREL_API_KEY = os.getenv("NREL_API_KEY")
set_developer_nrel_gov_key(NREL_API_KEY)  # Set this key manually here if you are not setting it using the .env


def establish_save_output_dict():
    """
    Establishes and returns a 'save_outputs_dict' dict
    for saving the relevant analysis variables for each site.
    """

    save_outputs_dict = dict()
    save_outputs_dict['Site Name'] = list()
    save_outputs_dict['Scenario Choice'] = list()
    save_outputs_dict['Site Lat'] = list()
    save_outputs_dict['Site Lon'] = list()
    save_outputs_dict['ATB Year'] = list()
    save_outputs_dict['Resource Year'] = list()
    save_outputs_dict['Critical Load Factor'] = list()
    save_outputs_dict['kW continuous load'] = list()
    save_outputs_dict['Useful Life'] = list()
    save_outputs_dict['PTC'] = list()
    save_outputs_dict['ITC'] = list()
    save_outputs_dict['Discount Rate'] = list()
    save_outputs_dict['Debt Equity'] = list()
    save_outputs_dict['Hub Height (m)'] = list()
    save_outputs_dict['Storage Enabled'] = list()
    save_outputs_dict['Wind Cost kW'] = list()
    save_outputs_dict['Solar Cost kW'] = list()
    save_outputs_dict['Storage Cost kW'] = list()
    save_outputs_dict['Storage Cost kWh'] = list()
    save_outputs_dict['Storage Hours'] = list()
    save_outputs_dict['Wind MW built'] = list()
    save_outputs_dict['Solar MW built'] = list()
    save_outputs_dict['Storage MW built'] = list()
    save_outputs_dict['Storage MWh built'] = list()
    save_outputs_dict['Electrolyzer Size'] = list()
    save_outputs_dict['Electrolyzer Total System Size'] = list()
    save_outputs_dict['Battery Can Grid Charge'] = list()
    save_outputs_dict['Grid Connected HOPP'] = list()
    save_outputs_dict['Built Interconnection Size'] = list()
    save_outputs_dict['REOpt Interconnection Size'] = list()
    save_outputs_dict['Total Installed Cost $(HOPP)'] = list()
    save_outputs_dict['Total Yearly Electrical Output'] = list()
    save_outputs_dict['LCOE'] = list()
    save_outputs_dict['Total Annual H2 production (kg)'] = list()
    save_outputs_dict['Gut-Check Cost/kg H2 (non-levelized, includes elec if used)'] = list()
    save_outputs_dict['Levelized Cost/kg H2 (lcoe using installed and operation costs)'] = list()
    save_outputs_dict['Levelized Cost/kg H2 (CF Method - using annual cashflows per technology)'] = list()
    # save_outputs_dict['Levelized Cost H2/kg (new method - with operational costs)'] = list()
    # save_outputs_dict['Levelized Cost H2/kg (new method - no operational costs)'] = list()
    # save_outputs_dict['Levelized cost of H2 (excl. electricity) (H2A)'] = list()
    # save_outputs_dict['Levelized H2 Elec Feedstock Cost/kg (HOPP)'] = list()
    # save_outputs_dict['Total H2 cost/kg (H2A)'] = list()
    # save_outputs_dict['H2 Elec Feedstock Cost/kg (HOPP) Net Cap Cost Method'] = list()
    save_outputs_dict['H2A scaled total install cost'] = list()
    save_outputs_dict['H2A scaled total install cost per kw'] = list()
    save_outputs_dict['REOpt Energy Shortfall'] = list()
    save_outputs_dict['REOpt Curtailment'] = list()
    save_outputs_dict['HOPP Total Generation'] = list()
    save_outputs_dict['Wind Capacity Factor'] = list()
    save_outputs_dict['HOPP Energy Shortfall'] = list()
    save_outputs_dict['HOPP Curtailment'] = list()
    save_outputs_dict['Battery Generation'] = list()
    save_outputs_dict['Electricity to Grid'] = list()

    return save_outputs_dict

def h2_main():
    """
    Runs a Hydrogen Levelized Cost (HLC/LCOH) analysis for scenarios contained in "default_h2_scenarios.csv"
    """


    # Step 1: Establish output structure and special inputs
    # save_all_runs = pd.DataFrame()
    save_outputs_dict = establish_save_output_dict()
    year = 2013
    sample_site['year'] = year
    useful_life = 40
    critical_load_factor_list = [1]
    run_reopt_flag = False
    custom_powercurve = True
    storage_used = True
    battery_can_grid_charge = False
    grid_connected_hopp = False
    interconnection_size_mw = 100
    electrolyzer_sizes = [100]

    # which plots to show
    plot_power_production = False
    plot_battery = False
    plot_grid = False
    plot_h2 = False
    plot_reopt = False

    # Step 2: Load scenarios from .csv and enumerate
    # scenarios_df = pd.read_csv('H2 Baseline Future Scenarios Test Refactor.csv')
    parent_path = os.path.abspath(os.path.dirname(__file__))
    scenarios_df = pd.read_csv(os.path.join(parent_path,'default_h2_scenarios_100MW.csv'))
    for electrolyzer_size in electrolyzer_sizes:
        for critical_load_factor in critical_load_factor_list:
            for i, scenario in scenarios_df.iterrows():
                # TODO: Make scenario_choice, lookup all other values from dataframe from csv.

                # TODO:
                # -Pass through rotor diameter to pySAM
                # -Add wind, solar, storage installed costs
                # -Fix "H2 H2 xxx" text
                # print(scenario)

                kw_continuous = electrolyzer_size*1000
                load = [kw_continuous for x in
                        range(0, 8760)]  # * (sin(x) + pi) Set desired/required load profile for plant

                scenario_choice = scenario['Scenario Name']
                site_name = scenario['Site Name']
                sample_site['lat'] = scenario['Lat']
                sample_site['lon'] = scenario['Long']
                lat = scenario['Lat']
                lon = scenario['Long']
                atb_year = scenario['ATB Year']
                ptc_avail = scenario['PTC Available']
                itc_avail = scenario['ITC Available']
                forced_sizes = scenario['Force Plant Size']
                force_electrolyzer_cost = scenario['Force Electrolyzer Cost']
                force_electrolyzer_spec = scenario['Elec Spec']
                elec_efficiency = scenario['Efficiency']
                if forced_sizes:
                    forced_wind_size = scenario['Wind Size MW']
                    forced_solar_size = scenario['Solar Size MW']
                    forced_storage_size_mw = scenario['Storage Size MW']
                    forced_storage_size_mwh = scenario['Storage Size MWh']
                else:
                    print("Using ReOPT for sizing. REopt will be turned on and may not find a solution")
                    run_reopt_flag = True

                if force_electrolyzer_cost:
                    forced_electrolyzer_cost = scenario['Electrolyzer Cost KW']

                tower_height = scenario['Tower Height']
                rotor_diameter = scenario['Rotor Diameter']
                turbine_rating = scenario['Turbine Rating']
                wind_cost_kw = scenario['Wind Cost KW']
                custom_powercurve_path = scenario['Powercurve File']
                solar_cost_kw = scenario['Solar Cost KW']
                storage_cost_kw = scenario['Storage Cost KW']
                storage_cost_kwh = scenario['Storage Cost KWh']
                debt_equity_split = scenario['Debt Equity']

                buy_price = scenario['Buy From Grid ($/kWh)']
                sell_price = scenario['Sell To Grid ($/kWh)']

                #Todo: Add useful life to .csv scenario input instead
                scenario['Useful Life'] = useful_life

                site = SiteInfo(sample_site, hub_height=tower_height)

                # Step 3: Set up REopt run
                # ------------------------- #

                wind_size_mw, solar_size_mw, storage_size_mw,\
                storage_size_mwh, storage_hours, reopt_results, REoptResultsDF = run_reopt(site, scenario, load,
                                                             interconnection_size_mw*1000,
                                                             critical_load_factor, useful_life,
                  battery_can_grid_charge, storage_used, run_reopt_flag)

                # Step 4: Set up HOPP run
                # ------------------------- #
                if forced_sizes:
                    solar_size_mw = forced_solar_size
                    wind_size_mw = forced_wind_size
                    storage_size_mw = forced_storage_size_mw
                    storage_size_mwh = forced_storage_size_mwh

                # TODO: Replace electrolyzer size with interconnection size after testing
                # technologies = {'solar': solar_size_mw,  # mw system capacity
                #                 'wind': wind_size_mw,  # mw system capacity
                #                 'collection_system': True}
                technologies = {'pv':
                                    {'system_capacity_kw': solar_size_mw * 1000},
                                'wind':
                                    {'num_turbines': np.floor(scenario['Wind Size MW'] / scenario['Turbine Rating']),
                                     'turbine_rating_kw': scenario['Turbine Rating']*1000,
                                     'hub_height': scenario['Tower Height'],
                                     'rotor_diameter': scenario['Rotor Diameter']},
                                'battery': {
                                    'system_capacity_kwh': storage_size_mwh * 1000,
                                    'system_capacity_kw': storage_size_mw * 1000
                                    }
                                }

                hybrid_plant, combined_hybrid_power_production_hopp, combined_hybrid_curtailment_hopp,\
                energy_shortfall_hopp, annual_energies, wind_plus_solar_npv, npvs, lcoe =  \
                    hopp_for_h2(site, scenario, technologies,
                                wind_size_mw, solar_size_mw, storage_size_mw, storage_size_mwh, storage_hours,
                    wind_cost_kw, solar_cost_kw, storage_cost_kw, storage_cost_kwh,
                    kw_continuous, load,
                    custom_powercurve,
                    electrolyzer_size, grid_connected_hopp)  #electrolyzer size is interconnection size

                wind_installed_cost = hybrid_plant.wind.total_installed_cost
                solar_installed_cost = hybrid_plant.pv.total_installed_cost
                hybrid_installed_cost = hybrid_plant.grid.total_installed_cost

                if plot_power_production:
                    plt.figure(figsize=(4,4))
                    plt.title("HOPP power production")
                    plt.plot(combined_hybrid_power_production_hopp[200:300],label="wind + pv")
                    plt.plot(energy_shortfall_hopp[200:300],label="shortfall")
                    plt.plot(combined_hybrid_curtailment_hopp[200:300],label="curtailment")
                    plt.plot(load[200:300],label="electrolyzer rating")
                    plt.xlabel("time (hour)")
                    plt.ylabel("power production")
                    # plt.ylim(0,250000)
                    plt.legend()
                    plt.tight_layout()
                    plt.show()

                # Step 5: Run Simple Dispatch Model
                # ------------------------- #

                bat_model = SimpleDispatch()
                bat_model.Nt = len(energy_shortfall_hopp)
                bat_model.curtailment = combined_hybrid_curtailment_hopp
                bat_model.shortfall = energy_shortfall_hopp

                bat_model.battery_storage = storage_size_mwh * 1000
                bat_model.charge_rate = storage_size_mw * 1000
                bat_model.discharge_rate = storage_size_mw * 1000

                battery_used, excess_energy, battery_SOC = bat_model.run()
                combined_pv_wind_storage_power_production_hopp = combined_hybrid_power_production_hopp + battery_used

                if plot_battery:
                    plt.figure(figsize=(7,4))
                    plt.subplot(121)
                    plt.plot(combined_hybrid_curtailment_hopp[200:300],label="curtailment")
                    plt.plot(energy_shortfall_hopp[200:300],label="shortfall")
                    plt.plot(battery_SOC[200:300],label="state of charge")
                    # plt.ylim(0,350000)
                    # plt.plot(excess_energy[200:300],label="excess")
                    plt.plot(battery_used[200:300],"--",label="battery used")
                    plt.legend()

                    plt.subplot(122)
                    plt.plot(combined_pv_wind_storage_power_production_hopp[200:300],label="wind+pv+storage")
                    plt.plot(combined_hybrid_power_production_hopp[200:300],"--",label="wind+pv")
                    plt.plot(load[200:300],"--",label="electrolyzer rating")
                    # plt.ylim(0,225000)

                    plt.legend()
                    plt.suptitle("battery dispatch")
                    plt.tight_layout()

                    plt.show()

                if plot_grid:
                    plt.plot(combined_pv_wind_storage_power_production_hopp[200:300],label="before buy from grid")

                sell_price = 0.01
                buy_price = 0.05

                # sell_price = False
                # buy_price = False

                if sell_price:
                    profit_from_selling_to_grid = np.sum(excess_energy)*sell_price
                else:
                    profit_from_selling_to_grid = 0.0

                #buy_price = True # if you want to force no buy from grid
                if buy_price:
                    cost_to_buy_from_grid = 0.00

                    for i in range(len(combined_pv_wind_storage_power_production_hopp)):
                        if combined_pv_wind_storage_power_production_hopp[i] < kw_continuous:
                            cost_to_buy_from_grid += (kw_continuous-combined_pv_wind_storage_power_production_hopp[i])*buy_price
                            combined_pv_wind_storage_power_production_hopp[i] = kw_continuous
                else:
                    cost_to_buy_from_grid = 0.00

                energy_to_electrolyzer = [x if x < kw_continuous else kw_continuous for x in combined_pv_wind_storage_power_production_hopp]

                if plot_grid:
                    plt.plot(combined_pv_wind_storage_power_production_hopp[200:300],"--",label="after buy from grid")
                    plt.plot(energy_to_electrolyzer[200:300],"--",label="energy to electrolyzer")
                    plt.legend()
                    plt.show()

                # Step 6: Run the Python H2A model
                # ------------------------- #
                #TODO: Refactor H2A model call
                # Should take as input (electrolyzer size, cost, electrical timeseries, total system electrical usage (kwh/kg),
                # Should give as ouptut (h2 costs by net cap cost, levelized, total_unit_cost of hydrogen etc)   )

                # electrical_generation_timeseries = combined_pv_wind_storage_power_production_hopp
                electrical_generation_timeseries = np.zeros_like(energy_to_electrolyzer)
                electrical_generation_timeseries[:] = energy_to_electrolyzer[:]

                # Parangat model
                adjusted_installed_cost = hybrid_plant.grid._financial_model.Outputs.adjusted_installed_cost
                #NB: adjusted_installed_cost does NOT include the electrolyzer cost
                useful_life = scenario['Useful Life']
                net_capital_costs = reopt_results['outputs']['Scenario']['Site'] \
                                    ['Financial']['net_capital_costs']

                # intalled costs:
                # hybrid_plant.grid._financial_model.costs

                # system_rating = electrolyzer_size
                system_rating = wind_size_mw + solar_size_mw
                H2_Results, H2A_Results = run_h2_PEM.run_h2_PEM(elec_efficiency,force_electrolyzer_spec,electrical_generation_timeseries,electrolyzer_size,
                                kw_continuous,forced_electrolyzer_cost,lcoe,adjusted_installed_cost,useful_life,
                                net_capital_costs)

                if plot_h2:
                    hydrogen_hourly_production = H2_Results['hydrogen_hourly_production']
                    plt.figure(figsize=(6,3))

                    plt.subplot(121)
                    plt.plot(electrical_generation_timeseries[200:300])
                    plt.ylim(0,max(electrical_generation_timeseries[200:300])*1.2)
                    plt.plot(load[200:300],label="electrolyzer rating")
                    plt.title("energy to electrolyzer")

                    plt.subplot(122)
                    plt.plot(hydrogen_hourly_production[200:300])
                    plt.ylim(0,max(hydrogen_hourly_production[200:300])*1.2)
                    plt.title("hydrogen production")

                    plt.tight_layout()
                    plt.show()


                # TEMPORARY CORRECTION FOR PEM EFFICIENCY.
                # # Convert H2 production from ~72.55kWh eff to 55.5kWh/kg
                H2_Results['hydrogen_annual_output'] = H2_Results['hydrogen_annual_output'] * 72.55/55.5

                # Step 6.5: Intermediate financial calculation
                total_elec_production = np.sum(electrical_generation_timeseries) #REMOVE
                total_hopp_installed_cost = hybrid_plant.grid._financial_model.SystemCosts.total_installed_cost
                total_electrolyzer_cost = H2A_Results['scaled_total_installed_cost']
                total_system_installed_cost = total_hopp_installed_cost + total_electrolyzer_cost
                annual_operating_cost_hopp = (wind_size_mw * 1000 * 42) + (solar_size_mw * 1000 * 13)
                annual_operating_cost_h2 = H2A_Results['Fixed O&M'] * H2_Results['hydrogen_annual_output']
                total_annual_operating_costs = annual_operating_cost_hopp + annual_operating_cost_h2 + cost_to_buy_from_grid - profit_from_selling_to_grid
                # h_lcoe_no_op_cost = lcoe_calc((H2_Results['hydrogen_annual_output']), total_system_installed_cost,
                #                    0, 0.07, useful_life)

                h_lcoe = lcoe_calc((H2_Results['hydrogen_annual_output']), total_system_installed_cost,
                                   total_annual_operating_costs, 0.07, useful_life)

                # Cashflow Financial Calculation (Not sure that this includes electrical prices)
                discount_rate = scenario['Discount Rate']
                cf_wind_annuals = hybrid_plant.wind._financial_model.Outputs.cf_annual_costs
                cf_solar_annuals = hybrid_plant.pv._financial_model.Outputs.cf_annual_costs
                cf_h2_annuals = H2A_Results['expenses_annual_cashflow'] # This might be unreliable.
                cf_df = pd.DataFrame([cf_wind_annuals, cf_solar_annuals, cf_h2_annuals[:len(cf_wind_annuals)]],['Wind', 'Solar', 'H2'])
                results_dir = Path(__file__).parent / 'results/'
                cf_df.to_csv(os.path.join(results_dir, "Annual Cashflows_{}_{}_{}_discount_{}.csv".format(site_name, scenario_choice, atb_year, discount_rate)))

                #NPVs of wind, solar, H2

                npv_wind_costs = npf.npv(discount_rate, cf_wind_annuals)
                npv_solar_costs = npf.npv(discount_rate, cf_solar_annuals)
                npv_h2_costs = npf.npv(discount_rate, cf_h2_annuals)
                npv_total_costs = npv_wind_costs+npv_solar_costs+npv_h2_costs
                LCOH_cf_method = -npv_total_costs / (H2_Results['hydrogen_annual_output'] * useful_life)
                financial_summary_df = pd.DataFrame([scenario['Useful Life'], scenario['Wind Cost KW'], scenario['Solar Cost KW'], forced_electrolyzer_cost,
                                                     scenario['Debt Equity'], atb_year, ptc_avail, itc_avail,
                                                     discount_rate, npv_wind_costs, npv_solar_costs, npv_h2_costs, LCOH_cf_method],
                                                    ['Useful Life', 'Wind Cost KW', 'Solar Cost KW', 'Electrolyzer Cost KW', 'Debt Equity',
                                                     'ATB Year', 'PTC available', 'ITC available', 'Discount Rate', 'NPV Wind Expenses', 'NPV Solar Expenses', 'NPV H2 Expenses', 'LCOH cf method'])
                financial_summary_df.to_csv(os.path.join(results_dir, 'Financial Summary.csv'))

                # Gut Check H2 calculation (non-levelized)
                total_installed_and_operational_lifetime_cost = total_system_installed_cost + (30 * total_annual_operating_costs)
                lifetime_h2_production = 30 * H2_Results['hydrogen_annual_output']
                gut_check_h2_cost_kg = total_installed_and_operational_lifetime_cost / lifetime_h2_production
                if buy_price:
                    #amount_to_plant=np.sum(energy_to_electrolyzer) #total energy to plant
                    amount_to_plant=total_elec_production
                    #amount_excess=np.sum(excess_energy)
                    #amount_battery=np.sum(battery_used)
                    amount_bought=cost_to_buy_from_grid/buy_price #energy from grid
                    #amount_bought_2=np.sum(energy_shortfall_hopp) #energy needed from grid/battery
                    #amount_produced=np.sum(hybrid_plant.grid.generation_profile)
                    #amount_HOPP_to_plant=amount_to_plant-amount_bought
                    percent_bought=amount_bought/amount_to_plant
                # Step 7: Print  results



                print_reults = True
                print_h2_results = True

                if print_reults:
                    # ------------------------- #
                    #TODO: Tidy up these print statements
                    print("Site Name: {}".format(scenario['Site Name']))
                    print("Future Scenario: {}".format(scenario['Scenario Name']))
                    #print("Wind Cost per KW: {}".format(scenario['Wind Cost KW']))
                    #print("PV Cost per KW: {}".format(scenario['Solar Cost KW']))
                    #print("Storage Cost per KW: {}".format(scenario['Storage Cost kW']))
                    #print("Storage Cost per KWh: {}".format(scenario['Storage Cost kWh']))
                    #print("Wind Size built: {}".format(wind_size_mw))
                    #print("PV Size built: {}".format(solar_size_mw))
                    #print("Storage Size built: {}".format(storage_size_mw))
                    #print("Storage Size built: {}".format(storage_size_mwh))
                    #print("Levelized cost of Electricity (HOPP): {}".format(lcoe))
                    print("Total Yearly Electrical Output (KW): {}".format(total_elec_production))
                   #print("Total Yearly Hydrogen Production: {}".format(H2_Results['hydrogen_annual_output']))
                    #print("Levelized Cost H2/kg (new method - no operational costs)".format(h_lcoe_no_op_cost))
                   # print("Capacity Factor of Electrolyzer: {}".format(H2_Results['cap_factor']))

                if print_h2_results:
                    #print('Total Lifetime H2(kg) produced: {}'.format(lifetime_h2_production))
                    #print("Gut-check H2 cost/kg: {}".format(gut_check_h2_cost_kg))
                    #print("h_lcoe: ", h_lcoe)
                    #print("LCOH CF Method (doesn't include elec)", LCOH_cf_method)
                    print("Total Yearly Hydrogen Production: {}".format(H2_Results['hydrogen_annual_output']))
                    #print("Levelized cost of Electricity $/kwh (HOPP): {}".format(lcoe))
                    #print("lcoe real",hybrid_plant.lcoe_real.hybrid)
                    #print("Lcoe nominal",hybrid_plant.lcoe_nom.hybrid)
                    # print("Levelized cost of H2 (electricity feedstock) (HOPP): {}".format(
                    #     H2_Results['feedstock_cost_h2_levelized_hopp']))
                    #print("Levelized cost of H2 (excl. electricity) (H2A): {}".format(H2A_Results['Total Hydrogen Cost ($/kgH2)']))
                    #print("Total unit cost of H2 ($/kg) : {}".format(H2_Results['total_unit_cost_of_hydrogen']))
                    # print("kg H2 cost from net cap cost/lifetime h2 production (HOPP): {}".format(
                    #     H2_Results['feedstock_cost_h2_via_net_cap_cost_lifetime_h2_hopp']))
                    #print(H2_Results['hydrogen_hourly_production'])
                    #print(H2_Results['hydrogen_annual_output'])


                Green_steel = True
                if Green_steel:
                    if buy_price:
                        buy_price=buy_price*100
                        lcoe=(lcoe*(1-percent_bought))+(buy_price*(percent_bought))

                    eta_el=.6
                    plant_life=40
                    tax_rate=.25
                    interest_rate=discount_rate
                    iron_ore_cost=90
                    emission_cost=0
                    carbon_steel_price=700
                    O2_price=40
                    el_spec=scenario['Elec Spec']
                    emission_factor=0
                    perc_O2_sold=.6
                    #electrolyzer_cost=force_electrolyzer_cost
                    electrolyzer_cost=scenario['Electrolyzer Cost KW'] #usd/kw
                    electricity_cost_USDMWH=lcoe*10
                    #True_electrolyzer_size=H2A_Results['total_plant_size']
                    True_electrolyzer_size=H2A_Results['total_plant_size']
                    import HAGS_Testing as HDRI
                    mass_h20, Green_steel_results=HDRI.HDRI_EAF_Model(scenario['Site Name'],scenario['Scenario Name'],eta_el,H2_Results['hydrogen_annual_output'],plant_life,tax_rate,interest_rate,electricity_cost_USDMWH,iron_ore_cost,emission_cost,carbon_steel_price,O2_price,el_spec,electrolyzer_cost/1000,emission_factor,total_elec_production,perc_O2_sold,True_electrolyzer_size)
                    #print(electrolyzer_cost)
                    #print("Levelized Cost of Steel ($/tonne): {}".format(Green_steel_results['Levelized Cost of Steel'][0]))
                    #print("NPV of Entire Green Steel Factory (MUSD/tonne): {}".format(Green_steel_results['Net Present Value'][0]))
                    #print("Levelized Cost of Hydrogen (Steel plant Electrolyzer): {}".format(Green_steel_results['Levelized Cost of Hydrogen'][0]))
                    #print("Mass H2O reentering Electrolyzer (kg): ",mass_h20)
                    #print("Electrical Demand of Steel Plant if H2 driven (kw): {}".format(Green_steel_results['Electrical Demand of Plant KW'][0]))
                    #print("Steel Produced per year in Tonnes per year: {}".format(Green_steel_results['Yearly Steel Production'][0]))
                    #print("Levelized Cost of Oxygen ($/kg): {}".format(Green_steel_results['Levelized Cost of Oxygen'][0]))
                    #print("Emissions in tonnes CO2: {}".format(Green_steel_results['Emissions'][0]))


                    #print(Green_steel_results)
                    file_name=scenario['Site Name']+' '+scenario['Scenario Name']+' '+'%s MW electrolyzer'% electrolyzer_size+' Buy from grid %s'% buy_price
                    Green_steel_results_df=pd.DataFrame.from_dict(Green_steel_results)
                    Green_steel_results_df.to_csv('C:/Users/CKIEFER/GreenSteel/HDRI-EAF-Technoeconomic-model/Hags_Outputs/%s.csv'% file_name)







                # Step 8: Plot REopt results

                if plot_reopt:
                    plot_reopt_results(REoptResultsDF, site_name, atb_year, critical_load_factor,
                                    useful_life, tower_height,
                                    wind_size_mw, solar_size_mw, storage_size_mw, storage_size_mwh, lcoe,
                                    H2_Results['feedstock_cost_h2_via_net_cap_cost_lifetime_h2_hopp'],
                                    H2_Results['feedstock_cost_h2_levelized_hopp'],
                                    hybrid_installed_cost, H2A_Results['Total Hydrogen Cost ($/kgH2)'],
                                    H2_Results['total_unit_cost_of_hydrogen'],
                                    output_dir='results/',
                                    monthly_separation=False, reopt_was_run=run_reopt_flag)

                # Step 9: Plot HOPP Production, Curtailment, and Hydrogen Production Profiles
                #TODO: Add this

                # Step 10: Save outputs
                # ------------------------- #
                #TODO: Place in function
                save_outputs_dict['Site Name'].append(site_name)
                save_outputs_dict['Scenario Choice'].append(scenario_choice)
                save_outputs_dict['Site Lat'].append(lat)
                save_outputs_dict['Site Lon'].append(lon)
                save_outputs_dict['ATB Year'].append(atb_year)
                save_outputs_dict['Resource Year'].append(year)
                save_outputs_dict['Critical Load Factor'].append(critical_load_factor)
                save_outputs_dict['kW continuous load'].append(kw_continuous)
                save_outputs_dict['Useful Life'].append(useful_life)
                save_outputs_dict['PTC'].append(ptc_avail)
                save_outputs_dict['ITC'].append(itc_avail)
                save_outputs_dict['Discount Rate'].append(discount_rate)
                save_outputs_dict['Debt Equity'].append(debt_equity_split)
                save_outputs_dict['Hub Height (m)'].append(tower_height)
                save_outputs_dict['Storage Enabled'].append(storage_used)
                save_outputs_dict['Wind Cost kW'].append(wind_cost_kw)
                save_outputs_dict['Solar Cost kW'].append(solar_cost_kw)
                save_outputs_dict['Storage Cost kW'].append(storage_cost_kw)
                save_outputs_dict['Storage Cost kWh'].append(storage_cost_kwh)
                save_outputs_dict['Storage Hours'].append(storage_hours)
                save_outputs_dict['Wind MW built'].append(wind_size_mw)
                save_outputs_dict['Solar MW built'].append(solar_size_mw)
                save_outputs_dict['Storage MW built'].append(storage_size_mw)
                save_outputs_dict['Storage MWh built'].append(storage_size_mwh)
                save_outputs_dict['Battery Can Grid Charge'].append(battery_can_grid_charge)
                save_outputs_dict['Built Interconnection Size'].append(hybrid_plant.interconnect_kw)
                save_outputs_dict['REOpt Interconnection Size'].append(interconnection_size_mw*1000)
                save_outputs_dict['Total Installed Cost $(HOPP)'].append(total_hopp_installed_cost)
                save_outputs_dict['Total Yearly Electrical Output'].append(total_elec_production)
                save_outputs_dict['LCOE'].append(lcoe)
                save_outputs_dict['Total Annual H2 production (kg)'].append(H2_Results['hydrogen_annual_output'])
                save_outputs_dict['Gut-Check Cost/kg H2 (non-levelized, includes elec if used)'].append(gut_check_h2_cost_kg)
                save_outputs_dict['Levelized Cost/kg H2 (lcoe using installed and operation costs)'].append(h_lcoe)
                save_outputs_dict['Levelized Cost/kg H2 (CF Method - using annual cashflows per technology)'].append(LCOH_cf_method)
                # save_outputs_dict['Levelized Cost H2/kg (new method - no operational costs)'].append(h_lcoe_no_op_cost)
                # save_outputs_dict['Levelized H2 Elec Feedstock Cost/kg (HOPP)'].append(H2_Results['feedstock_cost_h2_levelized_hopp'])
                # save_outputs_dict['Levelized cost of H2 (excl. electricity) (H2A)'].append(H2A_Results['Total Hydrogen Cost ($/kgH2)'])
                # save_outputs_dict['H2 Elec Feedstock Cost/kg (HOPP) Net Cap Cost Method'].append(H2_Results['feedstock_cost_h2_via_net_cap_cost_lifetime_h2_hopp'])
                # save_outputs_dict['Total H2 cost/kg (H2A)'].append(H2_Results['total_unit_cost_of_hydrogen'])
                save_outputs_dict['REOpt Energy Shortfall'].append(np.sum(REoptResultsDF['energy_shortfall']))
                save_outputs_dict['REOpt Curtailment'].append(np.sum(REoptResultsDF['combined_pv_wind_curtailment']))
                save_outputs_dict['Grid Connected HOPP'].append(grid_connected_hopp)
                save_outputs_dict['HOPP Total Generation'].append(np.sum(hybrid_plant.grid.generation_profile[0:8759]))
                save_outputs_dict['Wind Capacity Factor'].append(hybrid_plant.wind._system_model.Outputs.capacity_factor)
                save_outputs_dict['HOPP Energy Shortfall'].append(np.sum(energy_shortfall_hopp))
                save_outputs_dict['HOPP Curtailment'].append(np.sum(combined_hybrid_curtailment_hopp))
                save_outputs_dict['Battery Generation'].append(np.sum(battery_used))
                save_outputs_dict['Electricity to Grid'].append(np.sum(excess_energy))
                save_outputs_dict['Electrolyzer Size'].append(H2A_Results['electrolyzer_size'])
                save_outputs_dict['Electrolyzer Total System Size'].append(H2A_Results['total_plant_size'])
                save_outputs_dict['H2A scaled total install cost'].append(H2A_Results['scaled_total_installed_cost'])
                save_outputs_dict['H2A scaled total install cost per kw'].append(H2A_Results['scaled_total_installed_cost_kw'])


        # save_all_runs = save_all_runs.append(save_outputs_dict, sort=False)

    # Create dataframe from outputs and save

    save_outputs = False
    if save_outputs:
        save_outputs_dict_df = pd.DataFrame(save_outputs_dict)
        save_outputs_dict_df.to_csv(os.path.join(results_dir, "H2_Analysis_{}.csv".format('Main')))


if __name__ == "__main__":
   """
   Runs h2_main as a standalone analysis.
   Performs a Hydrogen Levelized Cost (HLC/LCOH) analysis for scenarios contained in "default_h2_scenarios.csv"
   Results are saved to ..results/H2_Analysis_Main.csv
   """
   h2_main()


import math

def calc_platform_opex(capex, opex_rate=0.011):
    '''
    Simple opex calculation based on a capex
        https://www.acm.nl/sites/default/files/documents/study-on-estimation-method-for-additional-efficient-offshore-grid-opex.pdf
    
    Output in $USD/year
    '''
    
    opex = capex * opex_rate    # USD/year
    
    return opex

def install_platform(mass, area, distance, install_duration=14, vessel=None, foundation="fixed"):
    '''
    A simplified platform installation costing model. 
    Total Cost = install_cost * duration 
         Compares the mass and/or deck space of equipment to the vessel limits to determine 
         the number of trips. Add an additional "at sea" install duration
          
    '''
    
    # If no ORBIT vessel is defined set default values (based on ORBIT's floating_heavy_lift_vessel)
    if vessel == None:
        if foundation=="fixed":
            # If no ORBIT vessel is defined set default values (based on ORBIT's example_heavy_lift_vessel)
            # Default values are from [3].
            vessel_cargo_mass = 7999 # t
            vessel_deck_space = 3999 # m**2 
            vessel_day_rate = 500001 # USD/day 
            vessel_speed = 5 # km/hr 
        elif foundation=="floating":
            # If no ORBIT vessel is defined set default values (based on ORBIT's floating_heavy_lift_vessel)
            vessel_cargo_mass = 7999 # t
            vessel_deck_space = 3999 # m**2 
            vessel_day_rate = 500001 # USD/day 
            vessel_speed = 7 # km/hr 
        else:
            raise(ValueError("Invalid offshore platform foundation type. Must be one of ['fixed', 'floating']"))
    else:
        vessel_cargo_mass = vessel.storage.max_cargo_mass # t
        vessel_deck_space = vessel.storage.max_deck_space # m**2 
        vessel_day_rate = vessel.day_rate # USD/day 
        vessel_speed = vessel.transit_speed # km/hr 

    # Get the # of trips based on ships cargo/space limits 
    num_of_trips = math.ceil(max((mass / vessel_cargo_mass), (area / vessel_deck_space)))

    # Total duration = double the trips + install_duration
    duration = (2 * num_of_trips * distance) / (vessel_speed * 24) + install_duration # days\

    # Final install cost is obtained by using the vessel's daily rate 
    install_cost = vessel_day_rate * duration   # USD

    return install_cost



import numpy as np
import pandas as pd
from greenheart.simulation.technologies.hydrogen.electrolysis.run_PEM_master import run_PEM_clusters
   
def clean_up_final_outputs(h2_tot,h2_ts):
  
   new_h2_tot = h2_tot.drop(['Cluster Rated H2 Production [kg/hr]','Cluster Rated Power Consumed [kWh]','Cluster Rated H2 Production [kg/yr]',\
      'Stack Rated H2 Production [kg/hr]','Stack Rated Power Consumed [kWh]'])
   h2_ts.sum(axis=1)
   ts_sum_desc = ['Input Power [kWh]','Power Consumed [kWh]',\
      'hydrogen production no start-up time','hydrogen_hourly_production',\
      'water_hourly_usage_kg']
   
   # new_h2_ts = h2_ts.drop(['V_cell With Deg','Power Per Stack [kW]','Stack Current [A]'])
   new_h2_ts = h2_ts.loc[ts_sum_desc].sum(axis=1)
   return new_h2_ts,new_h2_tot
   # return new_h2_ts,new_h2_tot
def combine_cluster_annual_performance_info(h2_tot):
   clusters = h2_tot.loc['Performance By Year'].index.to_list()
   performance_metrics = list(h2_tot.loc['Performance By Year'].iloc[0].keys())
   vals_to_sum = [k for k in performance_metrics if '/year' in k]
   n_years = len(h2_tot.loc['Performance By Year'].iloc[0][performance_metrics[0]].values())
   yr_keys = list(h2_tot.loc['Performance By Year'].iloc[0][performance_metrics[0]].keys())

   vals_to_average = [k for k in performance_metrics if '/year' not in k]
   new_dict = {}
   # for k in vals_to_sum:
   for k in performance_metrics:
      vals = np.zeros(n_years)
      for c in clusters:
         vals += np.array(list(h2_tot.loc['Performance By Year'].loc[c][k].values()))
         # vals += np.array(h2_tot.loc['Performance By Year'].loc[c][k].values())
      
      if k in vals_to_average:
         vals = vals/len(clusters)
      new_dict[k]= dict(zip(yr_keys,vals))
   return new_dict

def run_h2_PEM(electrical_generation_timeseries, 
               electrolyzer_size,
               useful_life, 
               n_pem_clusters,
               pem_control_type,
               electrolyzer_direct_cost_kw, 
               user_defined_pem_param_dictionary,
               grid_connection_scenario,
               hydrogen_production_capacity_required_kgphr,
               debug_mode = False,
               verbose=True):
   #last modified by Elenya Grant on 9/21/2023
   
   pem=run_PEM_clusters(electrical_generation_timeseries,electrolyzer_size,n_pem_clusters,electrolyzer_direct_cost_kw,useful_life,user_defined_pem_param_dictionary,verbose=verbose)

   if grid_connection_scenario!='off-grid':
      h2_ts,h2_tot=pem.run_grid_connected_pem(electrolyzer_size,hydrogen_production_capacity_required_kgphr)
   else:
      if pem_control_type == 'optimize':
         h2_ts,h2_tot=pem.run(optimize=True)
      else:
         h2_ts,h2_tot=pem.run()
   #dictionaries of performance during each year of simulation, 
   #good to use for a more accurate financial analysis
   annual_avg_performance = combine_cluster_annual_performance_info(h2_tot)
   
   #time-series info (unchanged)
   energy_input_to_electrolyzer=h2_ts.loc['Input Power [kWh]'].sum()
   hydrogen_hourly_production = h2_ts.loc['hydrogen_hourly_production'].sum()
   hourly_system_electrical_usage=h2_ts.loc['Power Consumed [kWh]'].sum()
   water_hourly_usage = h2_ts.loc['water_hourly_usage_kg'].sum()
   avg_eff_perc=39.41*hydrogen_hourly_production/hourly_system_electrical_usage
   hourly_efficiency=np.nan_to_num(avg_eff_perc)
   #simulation based average performance (unchanged)
   average_uptime_hr=h2_tot.loc['Total Uptime [sec]'].mean()/3600
   water_annual_usage = np.sum(water_hourly_usage)
   total_system_electrical_usage = np.sum(hourly_system_electrical_usage)
   tot_avg_eff=39.41/h2_tot.loc['Total kWh/kg'].mean()
   cap_factor_sim = h2_tot.loc['PEM Capacity Factor (simulation)'].mean()

   #Beginning of Life (BOL) Rated Specs (attributes/system design)
   max_h2_pr_hr = h2_tot.loc['Cluster Rated H2 Production [kg/hr]'].sum()
   max_pwr_pr_hr = h2_tot.loc['Cluster Rated Power Consumed [kWh]'].sum()
   rated_kWh_pr_kg = h2_tot.loc['Stack Rated Efficiency [kWh/kg]'].mean()
   elec_rated_h2_capacity_kgpy =h2_tot.loc['Cluster Rated H2 Production [kg/yr]'].sum()
   gal_h20_pr_kg_h2 = h2_tot.loc['gal H20 per kg H2'].mean()

   atrribute_desc = ["Efficiency [kWh/kg]","H2 Production [kg/hr]","Power Consumed [kWh]","Annual H2 Production [kg/year]",'Gal H2O per kg-H2']
   attribute_specs = ['Rated BOL: '+s for s in atrribute_desc]
   attributes = [rated_kWh_pr_kg,max_h2_pr_hr,max_pwr_pr_hr,elec_rated_h2_capacity_kgpy,gal_h20_pr_kg_h2]
   
   #Plant Life Average Performance
   system_avg_life_capfac = pd.Series(annual_avg_performance['Capacity Factor [-]']).mean()
   system_total_annual_h2_kg_pr_year = pd.Series(annual_avg_performance['Annual H2 Production [kg/year]']).mean()
   system_avg_life_eff_kWh_pr_kg =pd.Series(annual_avg_performance['Annual Average Efficiency [kWh/kg]']).mean()
   system_avg_life_eff_perc = pd.Series(annual_avg_performance['Annual Average Efficiency [%-HHV]']).mean()
   system_avg_life_energy_kWh_pr_yr = pd.Series(annual_avg_performance['Annual Energy Used [kWh/year]']).mean()
   
   average_stack_life_hrs = np.nanmean(h2_tot.loc['Stack Life [hours]'].values)
   average_time_until_replacement = np.nanmean(h2_tot.loc['Time until replacement [hours]'].values)
   life_vals = [system_avg_life_capfac,system_total_annual_h2_kg_pr_year,average_stack_life_hrs,average_time_until_replacement,system_avg_life_eff_kWh_pr_kg,system_avg_life_eff_perc,system_avg_life_energy_kWh_pr_yr]
   life_desc = ["Life: Capacity Factor","Life: Annual H2 production [kg/year]","Stack Life [hrs]","Time Until Replacement [hrs]","Life: Efficiency [kWh/kg]","Life: Efficiency [%-HHV]",'Life: Annual Power Consumption [kWh/year]']
   
   
   #Simulation Results
   sim = ["Capacity Factor","Active Time / Sim Time","Total Input Power [kWh]",\
      "Total H2 Produced [kg]",\
      "Average Efficiency [%-HHV]","Total Stack Off-Cycles","H2 Warm-Up Losses [kg]","Total Water Consumption [kg]"]
   
   sim_specs = ['Sim: '+s for s in sim]
   sim_performance = [cap_factor_sim, h2_tot.loc['Operational Time / Simulation Time (ratio)'].mean(),h2_tot.loc['Total Input Power [kWh]'].sum(),\
      h2_tot.loc['Total H2 Production [kg]'].sum(),\
      tot_avg_eff,h2_tot.loc['Total Off-Cycles'].sum(),h2_tot.loc['Warm-Up Losses on H2 Production'].sum(),water_annual_usage]
   
   
   H2_Results = dict(zip(attribute_specs,attributes))
   H2_Results.update(dict(zip(sim_specs,sim_performance)))
   H2_Results.update(dict(zip(life_desc,life_vals)))
   H2_Results.update({'Performance Schedules':pd.DataFrame(annual_avg_performance)})
   H2_Results.update({'Hydrogen Hourly Production [kg/hr]':hydrogen_hourly_production})
   H2_Results.update({'Water Hourly Consumption [kg/hr]':water_hourly_usage}) 
   
   if not debug_mode:
      h2_ts,h2_tot = clean_up_final_outputs(h2_tot,h2_ts)
  
   return H2_Results, h2_ts, h2_tot,energy_input_to_electrolyzer   
   

def run_h2_PEM_IVcurve(
      energy_to_electrolyzer,
      electrolyzer_size_mw,
      kw_continuous,
      electrolyzer_capex_kw,
      lcoe,
      adjusted_installed_cost,
      useful_life,
      net_capital_costs=0,
):
   
   # electrical_generation_timeseries = combined_pv_wind_storage_power_production_hopp
   electrical_generation_timeseries = np.zeros_like(energy_to_electrolyzer)
   electrical_generation_timeseries[:] = energy_to_electrolyzer[:]

   # system_rating = electrolyzer_size
   H2_Results, H2A_Results = kernel_PEM_IVcurve(
      electrical_generation_timeseries,
      electrolyzer_size_mw,
      useful_life,
      kw_continuous,
      electrolyzer_capex_kw,
      lcoe,
      adjusted_installed_cost,
      net_capital_costs)


   H2_Results['hydrogen_annual_output'] = H2_Results['hydrogen_annual_output']
   H2_Results['cap_factor'] = H2_Results['cap_factor']

   print("Total power input to electrolyzer: {}".format(np.sum(electrical_generation_timeseries)))
   print("Hydrogen Annual Output (kg): {}".format(H2_Results['hydrogen_annual_output']))
   print("Water Consumption (kg) Total: {}".format(H2_Results['water_annual_usage']))


   return H2_Results, H2A_Results # , electrical_generation_timeseries





import numpy as np
import numpy_financial as npf
from greenheart.to_organize.H2_Analysis.simple_cash_annuals import simple_cash_annuals
import warnings
from pytest import approx


def basic_H2_cost_model(electrolyzer_capex_kw, time_between_replacement,\
    electrolyzer_size_mw, useful_life, atb_year, 
    electrical_generation_timeseries_kw, hydrogen_annual_output, PTC_USD_kg, ITC_perc, include_refurb_in_opex=False, offshore=0):
    """
    Basic cost modeling for a PEM electrolyzer.
    Looking at cost projections for PEM electrolyzers over years 2022, 2025, 2030, 2035.
    Electricity costs are calculated outside of hydrogen cost model

    Needs:
    Scaling factor for off-shore electrolysis
    Verifying numbers are appropriate for simplified cash flows
    Verify how H2 PTC works/factors into cash flows

    If offshore = 1, then additional cost scaling is added to account for added difficulties for offshore installation, offshore=0 means onshore
    """

    # Basic information in our analysis
    discount_rate = 0.07
    kw_continuous = electrolyzer_size_mw *1000

    # Capacity factor
    avg_generation = np.mean(electrical_generation_timeseries_kw)  # Avg Generation
        # print("avg_generation: ", avg_generation)
    cap_factor = avg_generation / kw_continuous

    if cap_factor > 1.0:
        cap_factor = 1.0
        warnings.warn("Electrolyzer capacity factor would be greater than 1 with provided energy profile. Capacity factor has been reduced to 1 for electrolyzer cost estimate purposes.")
        
    # print(cap_factor)
    # if cap_factor != approx(1.0):
    #     raise(ValueError("Capacity factor must equal 1"))
    # print("cap_factor",cap_factor)

    # #Apply PEM Cost Estimates based on year based on GPRA pathway (H2New)
    # if atb_year == 2022:
    #     electrolyzer_capex_kw = 1100     #[$/kW capacity] stack capital cost
    #     time_between_replacement = 40000    #[hrs] 
    # elif atb_year == 2025:
    #     electrolyzer_capex_kw = 300
    #     time_between_replacement = 80000    #[hrs]
    # elif atb_year == 2030:
    #     electrolyzer_capex_kw = 150
    #     time_between_replacement = 80000    #[hrs]
    # elif atb_year == 2035:
    #     electrolyzer_capex_kw = 100
    #     time_between_replacement = 80000    #[hrs]

    # Hydrogen Production Cost From PEM Electrolysis - 2019 (HFTO Program Record)
    # https://www.hydrogen.energy.gov/pdfs/19009_h2_production_cost_pem_electrolysis_2019.pdf

    # Capital costs provide by Hydrogen Production Cost From PEM Electrolysis - 2019 (HFTO Program Record)
    stack_capital_cost = 342   #[$/kW]
    mechanical_bop_cost = 36  #[$/kW] for a compressor
    electrical_bop_cost = 82  #[$/kW] for a rectifier

    # Installed capital cost
    stack_installation_factor = 12/100  #[%] for stack cost 
    elec_installation_factor = 12/100   #[%] and electrical BOP

    # scale installation fraction if offshore (see Singlitico 2021 https://doi.org/10.1016/j.rset.2021.100005)
    stack_installation_factor *= 1 + offshore
    elec_installation_factor *= 1 + offshore

    #mechanical BOP install cost = 0%

    # Indirect capital cost as a percentage of installed capital cost
    site_prep = 2/100   #[%]
    engineering_design = 10/100 #[%]
    project_contingency = 15/100 #[%]
    permitting = 15/100     #[%]
    land = 250000   #[$]

    stack_replacment_cost = 15/100  #[% of installed capital cost]
    plant_lifetime = 40    #[years]
    fixed_OM = 0.24     #[$/kg H2]

    program_record = False

    # Chose to use numbers provided by GPRA pathways
    if program_record:
        total_direct_electrolyzer_cost_kw = (stack_capital_cost*(1 + stack_installation_factor)) \
            + mechanical_bop_cost + (electrical_bop_cost*(1 + elec_installation_factor))
    else:
        total_direct_electrolyzer_cost_kw = (electrolyzer_capex_kw * (1 + stack_installation_factor)) \
            + mechanical_bop_cost + (electrical_bop_cost*(1 + elec_installation_factor))

    # Assign CapEx for electrolyzer from capacity based installed CapEx
    electrolyzer_total_installed_capex = total_direct_electrolyzer_cost_kw*electrolyzer_size_mw*1000

    # Add indirect capital costs
    electrolyzer_total_capital_cost = ((site_prep + engineering_design + project_contingency + permitting)\
        *electrolyzer_total_installed_capex) + land + electrolyzer_total_installed_capex

    # O&M costs
    # https://www.sciencedirect.com/science/article/pii/S2542435121003068
    h2_FOM_kg = 0.24 #[$/kg] for 700 MW electrolyzer (https://www.hydrogen.energy.gov/pdfs/19009_h2_production_cost_pem_electrolysis_2019.pdf)
    scaled_h2_FOM_kg = h2_FOM_kg * electrolyzer_size_mw/700    # linearly scaled current central fixed O&M for a 700MW electrolyzer up to a 1000MW electrolyzer 
    h2_FOM_kWh = scaled_h2_FOM_kg / 55.5    #[$/kWh] used 55.5 kWh/kg for efficiency
    fixed_OM = h2_FOM_kWh * 8760 #[$/kW-y] 
    property_tax_insurance = 1.5/100    #[% of Cap/y]
    variable_OM = 1.30  #[$/MWh]

    # Amortized refurbishment expense [$/MWh]
    if not include_refurb_in_opex:
        amortized_refurbish_cost = 0.0
    else:
        amortized_refurbish_cost = (total_direct_electrolyzer_cost_kw*stack_replacment_cost)\
                *max(((useful_life*8760*cap_factor)/time_between_replacement-1),0)/useful_life/8760/cap_factor*1000

    # Total O&M costs [% of installed cap/year]
    total_OM_costs = ((fixed_OM+(property_tax_insurance*total_direct_electrolyzer_cost_kw))/total_direct_electrolyzer_cost_kw\
        +((variable_OM+amortized_refurbish_cost)/1000*8760*(cap_factor/total_direct_electrolyzer_cost_kw)))

    capacity_based_OM = True
    if capacity_based_OM:
        electrolyzer_OM_cost = electrolyzer_total_installed_capex * total_OM_costs     #Capacity based
    else:   
        electrolyzer_OM_cost = fixed_OM  * hydrogen_annual_output #Production based - likely not very accurate

    # Add in electrolyzer repair schedule (every 7 years)
    # Use if not using time between replacement given in hours 
    # Currently not added into further calculations
    electrolyzer_repair_schedule = []
    counter = 1
    for year in range(0,useful_life):
        if year == 0:
            electrolyzer_repair_schedule = np.append(electrolyzer_repair_schedule, [0])

        elif counter % time_between_replacement == 0:
            electrolyzer_repair_schedule = np.append(electrolyzer_repair_schedule, [1])

        else:
            electrolyzer_repair_schedule = np.append(electrolyzer_repair_schedule, [0])
        counter += 1
    electrolyzer_replacement_costs = electrolyzer_repair_schedule * (stack_replacment_cost* electrolyzer_total_installed_capex)
    # print("H2 replacement costs: ", electrolyzer_replacement_costs)

    # Include Hydrogen PTC from the Inflation Reduction Act (range $0.60 - $3/kg-H2)
    h2_tax_credit = [0] * useful_life
    h2_tax_credit[0:10] = [hydrogen_annual_output* PTC_USD_kg] * 10
    # print('H2 tax credit',h2_tax_credit)

    # Include ITC from IRA (range 0% - 50%)
    # ITC is expressed as a percentage of the total installed cost which reduces the annual tax liabiity in year one of the project cash flow.
    h2_itc = (ITC_perc/100) * electrolyzer_total_installed_capex
    cf_h2_itc = [0]*30
    cf_h2_itc[1] = h2_itc
    # print('ITC', cf_h2_itc)

    # Simple cash annuals
    cf_h2_annuals = - simple_cash_annuals(useful_life, useful_life, electrolyzer_total_capital_cost,\
        electrolyzer_OM_cost, 0.03)

    # print("CF H2 Annuals",cf_h2_annuals)

    # Add positive cashflow from tax credit
    cf_h2_annuals = np.add(cf_h2_annuals,h2_tax_credit)

    # print('Added H2 ptc with cash flows', cf_h2_annuals)

    #Add ITC
    cf_h2_annuals = np.add(cf_h2_itc,cf_h2_annuals)
    # print('Added H2 ITC with cash flows', cf_h2_annuals)

    return cf_h2_annuals, electrolyzer_total_capital_cost, electrolyzer_OM_cost, electrolyzer_capex_kw, time_between_replacement, h2_tax_credit, h2_itc

if __name__ == "__main__":

    import matplotlib.pyplot as plt
    # plot a sweep of sizes for OPEX and CAPEX

    electrolyzer_capex_kw = 1300 # $/kW
    time_between_replacement = 65000 # hours
    electrolyzer_sizes_mw = np.arange(1, 1000)
    useful_life = 30 # years
    atb_year = 2025
    # electrical_generation_timeseries_kw = np.sin(np.arange(0,24*365)*1E-3)*0.5E6 + 0.6E6
    hydrogen_annual_output = 0

    # for distributed
    ndivs = [2, 5, 10]

    opex = []
    capex = []
    opex_distributed = np.zeros((len(ndivs), len(electrolyzer_sizes_mw)))
    capex_distributed = np.zeros((len(ndivs), len(electrolyzer_sizes_mw)))

    for i, electrolyzer_size_mw in enumerate(electrolyzer_sizes_mw):

        electrical_generation_timeseries_kw = electrolyzer_size_mw*1000*np.ones(365*24)

        # centralized
        _, electrolyzer_total_capital_cost, electrolyzer_OM_cost, _, _, _, _ = basic_H2_cost_model(electrolyzer_capex_kw, time_between_replacement,\
            electrolyzer_size_mw, useful_life, atb_year, 
            electrical_generation_timeseries_kw, hydrogen_annual_output, 0, 0, include_refurb_in_opex=False, offshore=0)
        
        opex.append(electrolyzer_OM_cost)
        capex.append(electrolyzer_total_capital_cost)

        for j, div in enumerate(ndivs):

            # divided
            electrolyzer_size_mw_distributed = electrolyzer_size_mw/div
            electrical_generation_timeseries_kw_distibuted = electrical_generation_timeseries_kw/div

            _, electrolyzer_capital_cost_distributed, electrolyzer_OM_cost_distributed, electrolyzer_capex_kw_distributed, time_between_replacement, h2_tax_credit, h2_itc = basic_H2_cost_model(electrolyzer_capex_kw, time_between_replacement,\
                electrolyzer_size_mw_distributed, useful_life, atb_year, 
                electrical_generation_timeseries_kw_distibuted, hydrogen_annual_output, 0, 0, include_refurb_in_opex=False, offshore=0)
            # print(opex_distributed)
            opex_distributed[j, i] = electrolyzer_OM_cost_distributed*div
            capex_distributed[j, i] = electrolyzer_capital_cost_distributed*div
        
    fig, ax = plt.subplots(1,2, figsize=(6,3))
    ax[0].plot(electrolyzer_sizes_mw, np.asarray(capex)*1E-6, label="Centralized")
    ax[1].plot(electrolyzer_sizes_mw, np.asarray(opex)*1E-6, label="Centralized")

    for i, div in enumerate(ndivs):
        # dims(capex_distributed)
        ax[0].plot(electrolyzer_sizes_mw, np.asarray(capex_distributed[i])*1E-6, "--", label="%i Divisions" % (div))
        ax[1].plot(electrolyzer_sizes_mw, np.asarray(opex_distributed[i])*1E-6, "--", label="%i Divisions" % (div))

    ax[0].set(ylabel="CAPEX (M USD)", xlabel="Electrolyzer Size (MW)")
    ax[1].set(ylabel="Annual OPEX (M USD)", xlabel="Electrolyzer Size (MW)")
    plt.legend(frameon=False)
    plt.tight_layout()
    plt.show()

    ## plot divided energy signals
    fig, ax = plt.subplots(1)
    ax.plot(electrical_generation_timeseries_kw, label="%s" % (1))
    for i, div in enumerate(ndivs):
        ax.plot(electrical_generation_timeseries_kw/div, label="%s" % (div))

    ax.set(xlabel="Hour", ylabel="Power (MW)")
    plt.tight_layout()
    plt.show()




from pytest import approx, fixture, raises
import copy
from greenheart.simulation.technologies.steel import steel

ng_prices_dict = {
    "2035": 3.76232,
    "2036": 3.776032,
    "2037": 3.812906,
    "2038": 3.9107960000000004,
    "2039": 3.865776,
    "2040": 3.9617400000000003,
    "2041": 4.027136,
    "2042": 4.017166,
    "2043": 3.9715339999999997,
    "2044": 3.924314,
    "2045": 3.903287,
    "2046": 3.878192,
    "2047": 3.845413,
    "2048": 3.813366,
    "2049": 3.77735,
    "2050": 3.766164,
    "2051": 3.766164,
    "2052": 3.766164,
    "2053": 3.766164,
    "2054": 3.766164,
    "2055": 3.766164,
    "2056": 3.766164,
    "2057": 3.766164,
    "2058": 3.766164,
    "2059": 3.766164,
    "2060": 3.766164,
    "2061": 3.766164,
    "2062": 3.766164,
    "2063": 3.766164,
    "2064": 3.766164,
}
grid_prices_dict = {
    "2035": 89.42320514456621,
    "2036": 89.97947569251141,
    "2037": 90.53574624045662,
    "2038": 91.09201678840184,
    "2039": 91.64828733634704,
    "2040": 92.20455788429224,
    "2041": 89.87291235917809,
    "2042": 87.54126683406393,
    "2043": 85.20962130894978,
    "2044": 82.87797578383562,
    "2045": 80.54633025872147,
    "2046": 81.38632144593608,
    "2047": 82.22631263315068,
    "2048": 83.0663038203653,
    "2049": 83.90629500757991,
    "2050": 84.74628619479452,
    "2051": 84.74628619479452,
    "2052": 84.74628619479452,
    "2053": 84.74628619479452,
    "2054": 84.74628619479452,
    "2055": 84.74628619479452,
    "2056": 84.74628619479452,
    "2057": 84.74628619479452,
    "2058": 84.74628619479452,
    "2059": 84.74628619479452,
    "2060": 84.74628619479452,
    "2061": 84.74628619479452,
    "2062": 84.74628619479452,
    "2063": 84.74628619479452,
    "2064": 84.74628619479452,
}

financial_assumptions = {
    "total income tax rate": 0.2574,
    "capital gains tax rate": 0.15,
    "leverage after tax nominal discount rate": 0.10893,
    "debt equity ratio of initial financing": 0.624788,
    "debt interest rate": 0.050049,
}


@fixture
def cost_config():
    config = steel.SteelCostModelConfig(
        operational_year=2035,
        plant_capacity_mtpy=1084408.2137715619,
        lcoh=4.2986685034417045,
        feedstocks=steel.Feedstocks(
            natural_gas_prices=ng_prices_dict, oxygen_market_price=0
        ),
        o2_heat_integration=False,
    )
    return config


def test_run_steel_model():
    capacity = 100.0
    capacity_factor = 0.9

    steel_production_mtpy = steel.run_steel_model(capacity, capacity_factor)

    assert steel_production_mtpy == 90.0


def test_steel_cost_model(subtests, cost_config):

    res: steel.SteelCostModelOutputs = steel.run_steel_cost_model(cost_config)

    with subtests.test("CapEx"):
        assert res.total_plant_cost == approx(451513562.41513157)
    with subtests.test("Fixed OpEx"):
        assert res.total_fixed_operating_cost == approx(99119892.8431614)
    with subtests.test("Installation"):
        assert res.installation_cost == approx(179525207.856775)


def test_steel_finance_model(cost_config):
    # Parameter -> Hydrogen/Steel/Ammonia
    costs: steel.SteelCostModelOutputs = steel.run_steel_cost_model(cost_config)

    plant_capacity_factor = 0.9
    steel_production_mtpy = steel.run_steel_model(
        cost_config.plant_capacity_mtpy, plant_capacity_factor
    )

    config = steel.SteelFinanceModelConfig(
        plant_life=30,
        plant_capacity_mtpy=cost_config.plant_capacity_mtpy,
        plant_capacity_factor=plant_capacity_factor,
        steel_production_mtpy=steel_production_mtpy,
        lcoh=cost_config.lcoh,
        feedstocks=cost_config.feedstocks,
        grid_prices=grid_prices_dict,
        financial_assumptions=financial_assumptions,
        costs=costs,
    )

    lcos_expected = 961.2866791076059

    res: steel.SteelFinanceModelOutputs = steel.run_steel_finance_model(config)

    assert res.sol.get("price") == lcos_expected


def test_steel_size_h2_input(subtests):
    config = steel.SteelCapacityModelConfig(
        hydrogen_amount_kgpy=73288888.8888889,
        input_capacity_factor_estimate=0.9,
        feedstocks=steel.Feedstocks(
            natural_gas_prices=ng_prices_dict, oxygen_market_price=0
        ),
    )

    res: steel.SteelCapacityModelOutputs = steel.run_size_steel_plant_capacity(config)

    with subtests.test("steel plant size"):
        assert res.steel_plant_capacity_mtpy == approx(1000000)
    with subtests.test("hydrogen input"):
        assert res.hydrogen_amount_kgpy == approx(73288888.8888889)


def test_steel_size_steel_input(subtests):
    config = steel.SteelCapacityModelConfig(
        desired_steel_mtpy=1000000,
        input_capacity_factor_estimate=0.9,
        feedstocks=steel.Feedstocks(
            natural_gas_prices=ng_prices_dict, oxygen_market_price=0
        ),
    )

    res: steel.SteelCapacityModelOutputs = steel.run_size_steel_plant_capacity(config)

    with subtests.test("steel plant size"):
        assert res.steel_plant_capacity_mtpy == approx(1111111.111111111)
    with subtests.test("hydrogen input"):
        assert res.hydrogen_amount_kgpy == approx(73288888.8888889)


def test_run_steel_full_model(subtests):
    config = {
        "steel": {
            "capacity": {
                "input_capacity_factor_estimate": 0.9,
                "desired_steel_mtpy": 1000000,
            },
            "costs": {
                "operational_year": 2035,
                "o2_heat_integration": False,
                "feedstocks": {
                    "natural_gas_prices": ng_prices_dict,
                    "oxygen_market_price": 0,
                },
                "lcoh": 4.2986685034417045,
            },
            "finances": {
                "plant_life": 30,
                "lcoh": 4.2986685034417045,
                "grid_prices": grid_prices_dict,
                "financial_assumptions": financial_assumptions,
            },
        }
    }

    res = steel.run_steel_full_model(config)

    with subtests.test("output length"):
        assert len(res) == 3

    with subtests.test("h2 mass per year"):
        assert res[0].hydrogen_amount_kgpy == approx(73288888.8888889)
    
    with subtests.test("plant cost"):
        assert res[1].total_plant_cost == approx(458597254.6437372)
    
    with subtests.test("steel price"):
        assert res[2].sol.get("price") == approx(958.0597659101862)

def test_run_steel_full_model_changing_lcoh(subtests):
    config_0 = {
        "steel": {
            "capacity": {
                "input_capacity_factor_estimate": 0.9,
                "desired_steel_mtpy": 1000000,
            },
            "costs": {
                "operational_year": 2035,
                "o2_heat_integration": False,
                "feedstocks": {
                    "natural_gas_prices": ng_prices_dict,
                    "oxygen_market_price": 0,
                },
                "lcoh": 4.2986685034417045,
            },
            "finances": {
                "plant_life": 30,
                "lcoh": 4.2986685034417045,
                "grid_prices": grid_prices_dict,
                "financial_assumptions": financial_assumptions,
            },
        }
    }

    config_1 = copy.deepcopy(config_0)
    config_1["steel"]["costs"]["lcoh"] = 20.0
    config_1["steel"]["finances"]["lcoh"] = 20.0

    res0 = steel.run_steel_full_model(config_0)
    res1 = steel.run_steel_full_model(config_1)

    with subtests.test("output length 0"):
        assert len(res0) == 3
    with subtests.test("output length 1"):
        assert len(res1) == 3
    with subtests.test("res0 res1 equal h2 mass per year"):
        assert res0[0].hydrogen_amount_kgpy == res1[0].hydrogen_amount_kgpy
    with subtests.test("res0 res1 equal plant cost"):
        assert res0[1].total_plant_cost == res1[1].total_plant_cost
    with subtests.test("res0 price lt res1 price"):
        assert res0[2].sol.get("price") < res1[2].sol.get("price")
    with subtests.test("raise value error when LCOH values do not match"):
        with raises(ValueError, match="steel cost LCOH and steel finance LCOH are not equal"):
            config_1["steel"]["finances"]["lcoh"] = 40.0
            res1 = steel.run_steel_full_model(config_1)



"""
This file is based on the WISDEM file of the same name: https://github.com/WISDEM/WISDEM
"""

import os
import warnings

import numpy as np
from typing import Optional, Union

import openmdao.api as om

from greenheart.simulation.greenheart_simulation import GreenHeartSimulationConfig
from greenheart.tools.optimization.openmdao import TurbineDistanceComponent, BoundaryDistanceComponent
from hopp.simulation import HoppInterface

class PoseOptimization(object):
    """This class contains a collection of methods for setting up an openmdao optimization problem for a greenheart simulation.

    Args:
        config (GreenHeartSimulationConfig): instance of a greenheart config containing all desired simulation set up
    """
    def __init__(self, config: GreenHeartSimulationConfig):
        """This method primarily establishes lists of optimization methods available through different optimization drivers
        """

        self.config = config

        self.scipy_methods = [
            "SLSQP",
            "Nelder-Mead",
            "COBYLA",
        ]

        self.pyoptsparse_methods = [
            "SNOPT",
            "CONMIN",
            "NSGA2",
        ]

    def get_number_design_variables(self):
        """This method counts the number of design variables required given the provided set up and returns the result

        Returns:
            int: number of design variables
        """
        # Determine the number of design variables
        n_DV = 0
        
        if self.config.greenheart_config["opt_options"]["design_variables"]["electrolyzer_rating_kw"]["flag"]:
            n_DV += 1
        if self.config.greenheart_config["opt_options"]["design_variables"]["pv_capacity_kw"]["flag"]:
            n_DV += 1
        if self.config.greenheart_config["opt_options"]["design_variables"]["wave_capacity_kw"]["flag"]:
            n_DV += 1
        if self.config.greenheart_config["opt_options"]["design_variables"]["battery_capacity_kw"]["flag"]:
            n_DV += 1
        if self.config.greenheart_config["opt_options"]["design_variables"]["battery_capacity_kwh"]["flag"]:
            n_DV += 1
        if self.config.greenheart_config["opt_options"]["design_variables"]["turbine_x"]["flag"]:
            n_DV += self.config.hopp_config["technologies"]["wind"]["num_turbines"]
        if self.config.greenheart_config["opt_options"]["design_variables"]["turbine_y"]["flag"]:
            n_DV += self.config.hopp_config["technologies"]["wind"]["num_turbines"]
        
        # Wrap-up at end with multiplier for finite differencing
        if "form" in self.config.greenheart_config["opt_options"]["driver"]["optimization"].keys():
            if self.config.greenheart_config["opt_options"]["driver"]["optimization"]["form"] == "central": # TODO this should probably be handled at the MPI point to avoid confusion with n_DV being double what would be expected
                n_DV *= 2

        return n_DV

    def _get_step_size(self):
        """If a step size for the driver-level finite differencing is provided, use that step size. Otherwise use a default value.

        Returns:
            step size (float): step size for optimization
        """

        if not "step_size" in self.config.greenheart_config["opt_options"]["driver"]["optimization"]:
            step_size = 1.0E-6
            warnings.warn(f"Step size was not specified, setting step size to {step_size}. Step size may be set in the greenheart \
                          config file under opt_options/driver/optimization/step_size and should be of type float", UserWarning)
        else:
            step_size = self.config.greenheart_config["opt_options"]["driver"]["optimization"]["step_size"]

        return  step_size

    def _set_optimizer_properties(self, opt_prob, options_keys=[], opt_settings_keys=[], mapped_keys={}):
        """Set the optimizer properties, both the `driver.options` and
        `driver.opt_settings`. See OpenMDAO documentation on drivers
        to determine which settings are set by either options or
        opt_settings.

        Args:
            opt_prob (OpenMDAO problem object):  The hybrid plant OpenMDAO problem object.
            options_keys (list, optional): List of keys for driver opt_settings to be set. Defaults to [].
            opt_settings_keys (list, optional): List of keys for driver options to be set. Defaults to [].
            mapped_keys (dict, optional): Key pairs where the yaml name differs from what's expected
                                          by the driver. Specifically, the key is what's given in the yaml
                                          and the value is what's expected by the driver. Defaults to {}.

        Returns:
            opt_prob (OpenMDAO problem object): The updated openmdao problem object with driver settings applied.
        """

        opt_options = self.config.greenheart_config["opt_options"]["driver"]["optimization"]

        # Loop through all of the options provided and set them in the OM driver object
        for key in options_keys:
            if key in opt_options:
                if key in mapped_keys:
                    opt_prob.driver.options[mapped_keys[key]] = opt_options[key]
                else:
                    opt_prob.driver.options[key] = opt_options[key]

        # Loop through all of the opt_settings provided and set them in the OM driver object
        for key in opt_settings_keys:
            if key in opt_options:
                if key in mapped_keys:
                    opt_prob.driver.opt_settings[mapped_keys[key]] = opt_options[key]
                else:
                    opt_prob.driver.opt_settings[key] = opt_options[key]

        return opt_prob

    def set_driver(self, opt_prob):
        """set which optimization driver to use and set options

        Args:
            opt_prob (openmdao problem instance): openmdao problem class instance for current optimization problem

        Raises:
            ImportError: An optimization algorithm from pyoptsparse was selected, but pyoptsparse is not installed
            ImportError: An optimization algorithm from pyoptsparse was selected, but the algorithm code is not currently installed within pyoptsparse
            ImportError: An optimization algorithm was requested from NLopt, but NLopt is not currently installed.
            ValueError: The selected optimizer is not yet supported.
            Exception: The specified generator type for the OpenMDAO design of experiments is unsupported.

        Returns:
            opt_prob (openmdao problem instance): openmdao problem class instance, edited from input with desired driver and driver options
        """

        folder_output = self.config.greenheart_config["opt_options"]["general"]["folder_output"]

        if self.config.greenheart_config["opt_options"]["driver"]["optimization"]["flag"]:
            opt_options = self.config.greenheart_config["opt_options"]["driver"]["optimization"]
            step_size = self._get_step_size()

            if "step_calc" in opt_options.keys():
                if opt_options["step_calc"] == "None":
                    step_calc = None
                else:
                    step_calc = opt_options["step_calc"]
            else:
                step_calc = None

            if "form" in opt_options.keys():
                if opt_options["form"] == "None":
                    form = None
                else:
                    form = opt_options["form"]
            else:
                form = None

            opt_prob.model.approx_totals(method="fd", step=step_size, form=form, step_calc=step_calc)

            # Set optimization solver and options. First, Scipy's SLSQP and COBYLA
            if opt_options["solver"] in self.scipy_methods:
                opt_prob.driver = om.ScipyOptimizeDriver()
                opt_prob.driver.options["optimizer"] = opt_options["solver"]

                options_keys = ["tol", "max_iter", "disp"]
                opt_settings_keys = ["rhobeg", "catol", "adaptive"]
                mapped_keys = {"max_iter": "maxiter"}
                opt_prob = self._set_optimizer_properties(opt_prob, options_keys, opt_settings_keys, mapped_keys)

            # The next two optimization methods require pyOptSparse.
            elif opt_options["solver"] in self.pyoptsparse_methods:
                try:
                    from openmdao.api import pyOptSparseDriver
                except:
                    raise ImportError(
                        f"You requested the optimization solver {opt_options['solver']}, but you have not installed pyOptSparse. Please do so and rerun."
                    )
                opt_prob.driver = pyOptSparseDriver(gradient_method=opt_options["gradient_method"])
                
                try:
                    opt_prob.driver.options["optimizer"] = opt_options["solver"]
                except:
                    raise ImportError(
                        f"You requested the optimization solver {opt_options['solver']}, but you have not installed it within pyOptSparse. Please build {opt_options['solver']} and rerun."
                    )

                # Most of the pyOptSparse options have special syntax when setting them,
                # so here we set them by hand instead of using `_set_optimizer_properties` for SNOPT and CONMIN.
                if opt_options["solver"] == "CONMIN":
                    opt_prob.driver.opt_settings["ITMAX"] = opt_options["max_iter"]

                if opt_options["solver"] == "NSGA2":
                    opt_settings_keys = [
                        "PopSize",
                        "maxGen",
                        "pCross_real",
                        "pMut_real",
                        "eta_c",
                        "eta_m",
                        "pCross_bin",
                        "pMut_bin",
                        "PrintOut",
                        "seed",
                        "xinit",
                    ]
                    opt_prob = self._set_optimizer_properties(opt_prob, opt_settings_keys=opt_settings_keys)

                elif opt_options["solver"] == "SNOPT":
                    opt_prob.driver.opt_settings["Major optimality tolerance"] = float(opt_options["tol"])
                    opt_prob.driver.opt_settings["Major iterations limit"] = int(opt_options["max_major_iter"])
                    opt_prob.driver.opt_settings["Iterations limit"] = int(opt_options["max_minor_iter"])
                    opt_prob.driver.opt_settings["Major feasibility tolerance"] = float(opt_options["tol"])
                    if "time_limit" in opt_options:
                        opt_prob.driver.opt_settings["Time limit"] = int(opt_options["time_limit"])
                    opt_prob.driver.opt_settings["Summary file"] = os.path.join(folder_output, "SNOPT_Summary_file.txt")
                    opt_prob.driver.opt_settings["Print file"] = os.path.join(folder_output, "SNOPT_Print_file.txt")
                    if "hist_file_name" in opt_options:
                        opt_prob.driver.hist_file = opt_options["hist_file_name"]
                    if "verify_level" in opt_options:
                        opt_prob.driver.opt_settings["Verify level"] = opt_options["verify_level"]
                    else:
                        opt_prob.driver.opt_settings["Verify level"] = -1
                if "hotstart_file" in opt_options:
                    opt_prob.driver.hotstart_file = opt_options["hotstart_file"]

            elif opt_options["solver"] == "GA":
                opt_prob.driver = om.SimpleGADriver()
                options_keys = [
                    "Pc",
                    "Pm",
                    "bits",
                    "compute_pareto",
                    "cross_bits",
                    "elitism",
                    "gray",
                    "max_gen",
                    "multi_obj_exponent",
                    "multi_obj_weights",
                    "penalty_exponent",
                    "penalty_parameter",
                    "pop_size",
                    "procs_per_model",
                    "run_parallel",
                ]
                opt_prob = self._set_optimizer_properties(opt_prob, options_keys)

            else:
                raise ValueError(f"The {self.config.greenheart_config['opt_options']['driver']['optimization']['solver']} optimizer is not yet supported!")

            if opt_options["debug_print"]:
                opt_prob.driver.options["debug_print"] = ["desvars", "ln_cons", "nl_cons", "objs", "totals"]

        elif self.config.greenheart_config["opt_options"]["driver"]["design_of_experiments"]["flag"]:
            doe_options = self.config.greenheart_config["opt_options"]["driver"]["design_of_experiments"]
            if doe_options["generator"].lower() == "uniform":
                generator = om.UniformGenerator(
                    num_samples=int(doe_options["num_samples"]),
                    seed=doe_options["seed"],
                )
            elif doe_options["generator"].lower() == "fullfact":
                generator = om.FullFactorialGenerator(levels=int(doe_options["levels"]))
            elif doe_options["generator"].lower() == "plackettburman":
                generator = om.PlackettBurmanGenerator()
            elif doe_options["generator"].lower() == "boxbehnken":
                generator = om.BoxBehnkenGenerator()
            elif doe_options["generator"].lower() == "latinhypercube":
                generator = om.LatinHypercubeGenerator(
                    samples=int(doe_options["num_samples"]),
                    criterion=doe_options["criterion"],
                    seed=doe_options["seed"],
                )
            else:
                raise Exception("The generator type {} is unsupported.".format(doe_options["generator"]))

            # Initialize driver
            opt_prob.driver = om.DOEDriver(generator)

            if doe_options["debug_print"]:
                opt_prob.driver.options["debug_print"] = ["desvars", "ln_cons", "nl_cons", "objs"]

            # options
            opt_prob.driver.options["run_parallel"] = doe_options["run_parallel"]

        elif self.config.greenheart_config["opt_options"]["driver"]["step_size_study"]["flag"]:
            pass

        else:
            print("WARNING: Design variables are set to be optimized or studied, but no driver is selected.")
            print("         If you want to run an optimization, please enable a driver.")

        return opt_prob

    def set_objective(self, opt_prob):
        """Set merit figure. Each objective has its own scaling.  Check first for user override

        Args:
            opt_prob (openmdao problem instance): openmdao problem instance for current optimization problem

        Returns:
            opt_prob (openmdao problem instance): openmdao problem instance for current optimization problem with objective set
        """
        # 
        if self.config.greenheart_config["opt_options"]["merit_figure_user"]["name"] != "":
            coeff = -1.0 if self.config.greenheart_config["opt_options"]["merit_figure_user"]["max_flag"] else 1.0
            opt_prob.model.add_objective(self.config.greenheart_config["opt_options"]["merit_figure_user"]["name"],
                                       ref=coeff*np.abs(self.config.greenheart_config["opt_options"]["merit_figure_user"]["ref"]))

        return opt_prob

    def set_design_variables(self, opt_prob, config, hi):
        """Set optimization design variables.

        Args:
            opt_prob (openmdao problem instance): openmdao problem instance for current optimization problem
            config (GreenHeartSimulationConfig): data class containing modeling, simulation, and optimization settings
            hi (HoppInterface): Main HOPP class

        Returns:
            opt_prob (openmdao problem instance): openmdao problem instance for current optimization problem with design variables set
        """
        
        design_variables_dict = {}
        for key in self.config.greenheart_config["opt_options"]["design_variables"].keys():
            if self.config.greenheart_config["opt_options"]["design_variables"][key]["flag"]:
                design_variables_dict[key] = config.greenheart_config["opt_options"]["design_variables"][key]
            
        print("ADDING DESIGN VARIABLES:")
        for dv, d in design_variables_dict.items():
            print(f"   {dv}")
            opt_prob.model.add_design_var(dv, lower=d["lower"], upper=d["upper"], units=d["units"])

        return opt_prob
    
    def set_constraints(self, opt_prob, hi: Optional[Union[None, HoppInterface]] = None):
        """sets up optimization constraints for the greenheart optimization problem

        Args:
            opt_prob (openmdao problem instance): openmdao problem instance for current optimization problem
            hi (Optional[Union[None, HoppInterface]], optional): Accepts an instance of the HoppInterface class (required when x and y are design variables). Defaults to None.

        Raises:
            Exception: all design variables must have at least one of an upper and lower bound specified

        Returns:
            opt_prob (openmdao problem instance): openmdao problem instance for current optimization problem edited to include constraint setup
        """

        if (hi is not None) and (self.config.hopp_config["technologies"]["wind"]["model_name"] == "floris"):
            turbine_x_init = hi.system.wind.config.floris_config["farm"]["layout_x"]
            turbine_y_init = hi.system.wind.config.floris_config["farm"]["layout_y"]
        else:
            # randomly generate initial turbine locations if not provided
            turbine_x_init = 1E3*np.random.rand(self.config.hopp_config["technologies"]["wind"]["num_turbines"])
            turbine_y_init = 1E3*np.random.rand(self.config.hopp_config["technologies"]["wind"]["num_turbines"])

        # turbine spacing constraint
        if self.config.greenheart_config["opt_options"]["constraints"]["turbine_spacing"]["flag"]:
            lower = self.config.greenheart_config["opt_options"]["constraints"]["turbine_spacing"]["lower"]
            
            opt_prob.model.add_subsystem("con_spacing", subsys=TurbineDistanceComponent(turbine_x_init=turbine_x_init, turbine_y_init=turbine_y_init), promotes=["*"])
            opt_prob.model.add_constraint("spacing_vec", lower=lower)
        
        # bondary distance constraint
        if self.config.greenheart_config["opt_options"]["constraints"]["boundary_distance"]["flag"]:
            lower = self.config.greenheart_config["opt_options"]["constraints"]["boundary_distance"]["lower"]
            opt_prob.model.add_subsystem("con_boundary", subsys=BoundaryDistanceComponent(hopp_interface=self.config.greenheart_config, turbine_x_init=turbine_x_init, turbine_y_init=turbine_y_init), promotes=["*"])
            opt_prob.model.add_constraint("boundary_distance_vec", lower=lower)

        # solar/platform size
        if self.config.greenheart_config["opt_options"]["constraints"]["pv_to_platform_area_ratio"]["flag"]:
            upper = self.config.greenheart_config["opt_options"]["constraints"]["pv_to_platform_area_ratio"]["upper"]
            opt_prob.model.add_subsystem("con_pv_platform_area", subsys=om.ExecComp(['pv_platform_ratio=pv_area/platform_area']), promotes=["*"])
            opt_prob.model.add_constraint("pv_platform_ratio", upper=upper)

        # User constraints
        user_constr = self.config.greenheart_config["opt_options"]["constraints"]["user"]
        for k in range(len(user_constr)):
            var_k = user_constr[k]["name"]
            
            if "lower_bound" in user_constr[k]:
                lower_k = user_constr[k]["lower_bound"]
            elif "lower" in user_constr[k]:
                lower_k = user_constr[k]["lower"]
            else:
                lower_k = None
                
            if "upper_bound" in user_constr[k]:
                upper_k = user_constr[k]["upper_bound"]
            elif "upper" in user_constr[k]:
                lower_k = user_constr[k]["upper"]
            else:
                upper_k = None
                
            if "indices" in user_constr[k]:
                idx_k = user_constr[k]["indices"]
            else:
                idx_k = None

            if lower_k is None and upper_k is None:
                raise Exception(f"Must include a lower_bound and/or an upper bound for {var_k}")
            
            opt_prob.model.add_constraint(var_k, lower=lower_k, upper=upper_k, indices=idx_k)
        
        return opt_prob

    def set_recorders(self, opt_prob):
        """sets up a recorder for the openmdao problem as desired in the input yaml

        Args:
            opt_prob (openmdao problem instance): openmdao problem instance for current optimization problem

        Returns:
            opt_prob (openmdao problem instance): openmdao problem instance for current optimization problem edited to include a set up recorder
        """
        folder_output = self.config.greenheart_config["opt_options"]["general"]["folder_output"]

        # Set recorder on the OpenMDAO driver level using the `optimization_log`
        # filename supplied in the optimization yaml
        if self.config.greenheart_config["opt_options"]["recorder"]["flag"]:
            recorder = om.SqliteRecorder(os.path.join(folder_output, self.config.greenheart_config["opt_options"]["recorder"]["file_name"]))
            opt_prob.driver.add_recorder(recorder)
            opt_prob.add_recorder(recorder)

            opt_prob.driver.recording_options["excludes"] = ["*_df"]
            opt_prob.driver.recording_options["record_constraints"] = True
            opt_prob.driver.recording_options["record_desvars"] = True
            opt_prob.driver.recording_options["record_objectives"] = True

            if self.config.greenheart_config["opt_options"]["recorder"]["includes"]:
                opt_prob.driver.recording_options["includes"] = self.config.greenheart_config["opt_options"]["recorder"]["includes"]

        return opt_prob

    def set_initial(self, opt_prob, config):
        
        return opt_prob

    def set_restart(self, opt_prob):
        """prepares to restart from last recorded iteration if the original problem was set up for warm start

        Args:
            opt_prob (openmdao problem instance): openmdao problem instance for current optimization problem

        Returns:
            opt_prob (openmdao problem instance): openmdao problem instance for current optimization problem set up for warm start
        """

        if "warmstart_file" in self.config.greenheart_config["opt_options"]["driver"]["optimization"]:
            # Directly read the pyoptsparse sqlite db file
            from pyoptsparse import SqliteDict

            db = SqliteDict(self.config.greenheart_config["opt_options"]["driver"]["optimization"]["warmstart_file"])

            # Grab the last iteration's design variables
            last_key = db["last"]
            desvars = db[last_key]["xuser"]

            # Obtain the already-setup OM problem's design variables
            if opt_prob.model._static_mode:
                design_vars = opt_prob.model._static_design_vars
            else:
                design_vars = opt_prob.model._design_vars

            # Get the absolute names from the promoted names within the OM model.
            # We need this because the pyoptsparse db has the absolute names for
            # variables but the OM model uses the promoted names.
            prom2abs = opt_prob.model._var_allprocs_prom2abs_list["output"]
            abs2prom = {}
            for key in design_vars:
                abs2prom[prom2abs[key][0]] = key

            # Loop through each design variable
            for key in desvars:
                prom_key = abs2prom[key]

                # Scale each DV based on the OM scaling from the problem.
                # This assumes we're running the same problem with the same scaling
                scaler = design_vars[prom_key]["scaler"]
                adder = design_vars[prom_key]["adder"]

                if scaler is None:
                    scaler = 1.0
                if adder is None:
                    adder = 0.0

                scaled_dv = desvars[key] / scaler - adder

        return opt_prob


# -*- coding: utf-8 -*-
"""
Created on Sat Dec  3 10:02:07 2022

@author: ereznic2
"""

# import sys
import numpy as np
import pandas as pd

# # Specify file path to PyFAST
# sys.path.append('../PyFAST/')

# import src.PyFAST as PyFAST

import ProFAST

import os.path

def run_profast_for_h2_transmission(project_dir, max_hydrogen_production_rate_kg_hr,max_hydrogen_delivery_rate_kg_hr,pipeline_length_km,electrolyzer_capacity_factor,enduse_capacity_factor,before_after_storage,plant_life,elec_price,grid_prices_interpolated_USDperkWh,site_name,atb_year):

    dir0 = os.path.join(project_dir, "H2_Analysis")

    # max_hydrogen_production_rate_kg_hr = 14852.8
    # max_hydrogen_delivery_rate_kg_hr = 6023.84
    # pipeline_length_km = 50
    # electrolyzer_capacity_factor = 0.33
    # enduse_capacity_factor = 0.9
    # before_after_storage = 'before'
    # plant_life = 30
    # lcoe = 4.7

    # Nameplate capacity of transmission
    if before_after_storage == 'before':
        #hydrogen_flow_capacity_kg_day = max_hydrogen_production_rate_kg_hr*24
        hydrogen_flow_capacity_kg_yr = max_hydrogen_production_rate_kg_hr*8760
        transmission_capacity_factor = electrolyzer_capacity_factor
    if before_after_storage == 'after':
        #hydrogen_flow_capacity_kg_day = max_hydrogen_delivery_rate_kg_hr*24
        hydrogen_flow_capacity_kg_yr =  max_hydrogen_delivery_rate_kg_hr*8760
        transmission_capacity_factor = enduse_capacity_factor

    hydrogen_flow_capacity_kg_day = hydrogen_flow_capacity_kg_yr/365

    # Read in compressor and pipeline scaling csv
    pipeline_compressor_cost_data = pd.read_csv(os.path.join(dir0, "HDSAM4_compressor_costs.csv"),index_col = None,header = 0)
    pipeline_pipe_cost_data = pd.read_csv(os.path.join(dir0, "HDSAM4_pipecapex_USDperkm.csv"),index_col = None,header = 0)
    pipeline_pipe_fixedopex_data = pd.read_csv(os.path.join(dir0, "HDSAM4_pipefixedopex_percentperyear.csv"),index_col = None,header = 0)

    # Select region for pipe capex and FOM
    regions={'IN':'Great Lakes','TX':'Southwest','IA':'Great Plains','MS':'Southeast','MN':'Great Plains','WY':'Rocky Mountains'}
    region=regions[site_name]

    # Downselect compressor data
    if atb_year <=2025:
        pipeline_compressor_cost_data = pipeline_compressor_cost_data.loc[pipeline_compressor_cost_data['Production Volume']=='Low']
    else:
        pipeline_compressor_cost_data = pipeline_compressor_cost_data.loc[pipeline_compressor_cost_data['Production Volume']=='Mid']

    # Interpolate compressor and pipe costs based on capacity. Note that it would be preferable to interpolate the scaling factor and then apply the scaling factor,
    # but this will take more time. Could do it at some point
    pipe_capex_USDperkm =  np.interp(hydrogen_flow_capacity_kg_yr,pipeline_pipe_cost_data['Capacity (kg/yr)'].to_numpy(),pipeline_pipe_cost_data[region].to_numpy())
    pipeline_capex = pipe_capex_USDperkm*pipeline_length_km
    pipe_fixedopex_percentperyear=  np.interp(hydrogen_flow_capacity_kg_yr,pipeline_pipe_fixedopex_data['Capacity (kg/yr)'].to_numpy(),pipeline_pipe_fixedopex_data[region].to_numpy())
    pipeline_FOM_USD_yr = pipe_fixedopex_percentperyear*pipeline_capex
    compressor_capex = np.interp(hydrogen_flow_capacity_kg_yr,pipeline_compressor_cost_data['Nameplate Capacity [kg/yr]'].to_numpy(),pipeline_compressor_cost_data['Capital Cost [$]'].to_numpy())
    compressor_FOM_frac = np.interp(hydrogen_flow_capacity_kg_yr,pipeline_compressor_cost_data['Nameplate Capacity [kg/yr]'].to_numpy(),pipeline_compressor_cost_data['Fixed Operating Cost [fraction of CapCost/y]'].to_numpy())
    compressor_FOM_USD_yr = compressor_FOM_frac*compressor_capex

    compressor_energy_usage_kwhperkw = np.interp(hydrogen_flow_capacity_kg_yr,pipeline_compressor_cost_data['Nameplate Capacity [kg/yr]'].to_numpy(),pipeline_compressor_cost_data['Electricity Use (kWh/kg)'].to_numpy())

    # pipeline_capex_perkm = np.interp(hydrogen_flow_capacity_kg_day,pipeline_cost_data['Nameplate kg/d'].to_numpy(),pipeline_cost_data['Capital Cost [$/km]'].to_numpy())
    # pipeline_capex = pipeline_capex_perkm*pipeline_length_km
    # pipeline_FOM_frac = np.interp(hydrogen_flow_capacity_kg_day,pipeline_cost_data['Nameplate kg/d'].to_numpy(),pipeline_cost_data['Fixed Operating Cost [fraction of OvernightCapCost/y]'].to_numpy())
    # pipeline_FOM_USD_yr = pipeline_FOM_frac*pipeline_capex

    financial_assumptions = pd.read_csv('H2_Analysis/financial_inputs.csv',index_col=None,header=0)
    financial_assumptions.set_index(["Parameter"], inplace = True)
    financial_assumptions = financial_assumptions['Hydrogen/Steel/Ammonia']

    # Set up ProFAST
    pf = ProFAST.ProFAST()

    install_years = 3
    analysis_start = list(grid_prices_interpolated_USDperkWh.keys())[0] - install_years

    # Fill these in - can have most of them as 0 also
    gen_inflation = 0.00
    pf.set_params('commodity',{"name":'Hydrogen',"unit":"kg","initial price":100,"escalation":gen_inflation})
    pf.set_params('capacity',hydrogen_flow_capacity_kg_day) #units/day
    pf.set_params('maintenance',{"value":0,"escalation":gen_inflation})
    pf.set_params('analysis start year',analysis_start)
    pf.set_params('operating life',plant_life)
    pf.set_params('installation months',12*install_years)
    pf.set_params('installation cost',{"value":0,"depr type":"Straight line","depr period":4,"depreciable":False})
    pf.set_params('non depr assets',0)
    pf.set_params('end of proj sale non depr assets',0)
    pf.set_params('demand rampup',0)
    pf.set_params('long term utilization',transmission_capacity_factor)
    pf.set_params('credit card fees',0)
    pf.set_params('sales tax',0)
    pf.set_params('license and permit',{'value':00,'escalation':gen_inflation})
    pf.set_params('rent',{'value':0,'escalation':gen_inflation})
    pf.set_params('property tax and insurance',0)
    pf.set_params('admin expense',0)
    pf.set_params('total income tax rate',financial_assumptions['total income tax rate'])
    pf.set_params('capital gains tax rate',financial_assumptions['capital gains tax rate'])
    pf.set_params('sell undepreciated cap',True)
    pf.set_params('tax losses monetized',True)
    pf.set_params('general inflation rate',gen_inflation)
    pf.set_params('leverage after tax nominal discount rate',financial_assumptions['leverage after tax nominal discount rate'])
    pf.set_params('debt equity ratio of initial financing',financial_assumptions['debt equity ratio of initial financing'])
    pf.set_params('debt type','Revolving debt')
    pf.set_params('debt interest rate',financial_assumptions['debt interest rate'])
    pf.set_params('cash onhand',1)

    #----------------------------------- Add capital items to ProFAST ----------------
    pf.add_capital_item(name="Pipeline",cost=pipeline_capex,depr_type="MACRS",depr_period=7,refurb=[0])
    pf.add_capital_item(name="Compressor",cost=compressor_capex,depr_type="MACRS",depr_period=7,refurb=[0])

    total_capex = pipeline_capex + compressor_capex

    #-------------------------------------- Add fixed costs--------------------------------
    pf.add_fixed_cost(name="Pipeline Fixed O&M Cost",usage=1.0,unit='$/year',cost=pipeline_FOM_USD_yr,escalation=gen_inflation)
    pf.add_fixed_cost(name="Compressor Fixed O&M Cost",usage=1.0,unit='$/year',cost=compressor_FOM_USD_yr,escalation=gen_inflation)

    #---------------------- Add feedstocks, note the various cost options-------------------
    pf.add_feedstock(name='Electricity',usage=compressor_energy_usage_kwhperkw,unit='kWh',cost=grid_prices_interpolated_USDperkWh,escalation=gen_inflation)

    sol = pf.solve_price()

    summary = pf.summary_vals

    price_breakdown = pf.get_cost_breakdown()

    price_breakdown_compressor = price_breakdown.loc[price_breakdown['Name']=='Compressor','NPV'].tolist()[0]
    price_breakdown_pipeline = price_breakdown.loc[price_breakdown['Name']=='Pipeline','NPV'].tolist()[0]
    price_breakdown_compressor_FOM = price_breakdown.loc[price_breakdown['Name']=='Compressor Fixed O&M Cost','NPV'].tolist()[0]
    price_breakdown_pipeline_FOM = price_breakdown.loc[price_breakdown['Name']=='Pipeline Fixed O&M Cost','NPV'].tolist()[0]
    price_breakdown_electricity = price_breakdown.loc[price_breakdown['Name']=='Electricity','NPV'].tolist()[0]

    price_breakdown_taxes = price_breakdown.loc[price_breakdown['Name']=='Income taxes payable','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name'] == 'Monetized tax losses','NPV'].tolist()[0]

    price_breakdown_financial = price_breakdown.loc[price_breakdown['Name']=='Cash on hand reserve','NPV'].tolist()[0]\
        + price_breakdown.loc[price_breakdown['Name']=='Repayment of debt','NPV'].tolist()[0]\
        + price_breakdown.loc[price_breakdown['Name']=='Interest expense','NPV'].tolist()[0]\
        + price_breakdown.loc[price_breakdown['Name']=='Dividends paid','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name']=='Cash on hand recovery','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name']=='Inflow of debt','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name']=='Inflow of equity','NPV'].tolist()[0]

    price_check = price_breakdown_compressor+price_breakdown_pipeline+price_breakdown_compressor_FOM\
               + price_breakdown_pipeline_FOM+price_breakdown_electricity\
               + price_breakdown_taxes+price_breakdown_financial

    h2_transmission_price_breakdown = {'LCOHT: Compressor ($/kg)':price_breakdown_compressor,'LCOHT: Pipeline ($/kg)':price_breakdown_pipeline,\
                      'LCOHT: Compressor FOM ($/kg)':price_breakdown_compressor_FOM,'LCOHT: Pipeline FOM ($/kg)':price_breakdown_pipeline_FOM,\
                      'LCOHT: Compressor electricity ($/kg)':price_breakdown_electricity,'LCOHT: Taxes ($/kg)':price_breakdown_taxes,\
                      'LCOHT: Finances ($/kg)':price_breakdown_financial,'LCOHT: total ($/kg)':price_check}

    return(sol,summary,h2_transmission_price_breakdown,total_capex)



import pytest
import os

from greenheart.simulation.technologies.hydrogen.electrolysis.pem_mass_and_footprint import mass, footprint


def test_footprint_0mw():

    assert footprint(0.0) == 0.0

def test_footprint_1mw():

    assert footprint(1) == 48

def test_mass():

    assert mass(0.045) == pytest.approx(900.0, rel=1E-4)